"""Tests dict input objects for `tackle.providers.system.hooks.variable` module."""
from tackle.main import tackle


def test_provider_system_hook_var(change_dir):
    """Verify the hook call works properly."""
    output = tackle('tackle.yaml')

    assert output['short'] == output['a_list']


def test_provider_system_hook_var_lists_embed(change_dir):
    """Verify the hook call works properly."""
    output = tackle('lists.yaml')

    assert output['list_of_lists'] == [[1, 2], [3, 4]]



"""test for docs provider."""
# from tackle import tackle
#
#
# def test_provider_docs_provider_docs(change_dir):
#     """Run the provider docs."""
#     output = tackle('.')
#     assert output
#
#
# def test_provider_docs_provider_docs_requests(change_dir):
#     """Verify gcp get azs of region."""
#     output = tackle("requests.yaml", no_input=True)
#     assert output



"""Block hook."""
from tackle.models import BaseHook, Context, Field
from tackle.parser import walk_sync


class BlockHook(BaseHook):
    """
    Hook for blocks of hooks.

    This is a special case where the hooks input variables are not rendered
    until it is later executed. Each `item` is looped over and parsed like a
    normal pass. Useful if you have a block of hooks that should be grouped
    into a single conditional / looped execution.

    Render context is a little different than normal where both the context
    from outside of the hook and within the hook are made available. For
    instance in this contrived example:

    ```
    stuff: things
    block->:
      merge: true
      stuff->: print other_things
      things->: print "{{ stuff }}" --if "{{ stuff == 'things' }}"
    ```

    The output would be:

    ```
    stuff: other_things
    ```

    Because the higher level `stuff` takes precidance where as


    :param items: Map of inputs
    """

    hook_type: str = 'block'
    items: dict = Field(..., description="Items to be parsed like a normal input.")

    _render_exclude = {'items'}

    def execute(self):
        existing_context = self.output_dict.copy()
        existing_context.update(self.existing_context)

        tmp_context = Context(
            providers=self.providers,
            existing_context=existing_context,
            output_dict={},
            input_dict=self.items,
            key_path=[],
            no_input=self.no_input,
            calling_directory=self.calling_directory,
        )
        walk_sync(context=tmp_context, element=self.items.copy())
        return tmp_context.output_dict



"""Path hooks."""
from typing import Any
import logging
import os

from tackle.models import BaseHook, Field
from tackle.utils.paths import find_in_parent

logger = logging.getLogger(__name__)


class PathExistsListHook(BaseHook):
    """Hook for os package 'path.exists'."""

    hook_type: str = 'path_exists'
    path: str = Field(..., description="The path to file or directory")

    _args: list = ['path']

    def execute(self) -> bool:
        return os.path.exists(self.path)


class PathIsDirListHook(BaseHook):
    """Hook for os package 'path.isdir'."""

    hook_type: str = 'isdir'
    path: str = Field(..., description="The path to a directory")

    _args: list = ['path']

    def execute(self) -> bool:
        return os.path.isdir(self.path)


class PathIsFileListHook(BaseHook):
    """Hook for os package 'path.isfile'."""

    hook_type: str = 'isfile'
    path: str = Field(..., description="The path to a file")

    _args: list = ['path']

    def execute(self) -> bool:
        return os.path.isfile(self.path)


class FindInParentHook(BaseHook):
    """
    Hook to find the absolute path to a file or directory in parent directories.

    :return: string: Absolute path to the target file
    """

    hook_type: str = 'find_in_parent'
    target: str = Field(
        ..., description="The name of the file to find the absolute path to"
    )
    fallback: Any = Field(
        None, description="String to fallback on if the target is not found."
    )
    starting_dir: str = Field(
        '.',
        description="The starting directory to search from. Defaults to current working directory.",
    )

    _args: list = ['target']

    def execute(self):
        return find_in_parent(
            dir=self.starting_dir, targets=[self.target], fallback=self.fallback
        )


class FindInChildHook(BaseHook):
    """
    Hook to find the absolute path to a file or directory in child directories.

    :return: string: Absolute path to the target file
    """

    hook_type: str = 'find_in_child'
    target: str = Field(
        ..., description="The name of the file to find the absolute path to"
    )
    fallback: Any = Field(
        None, description="String to fallback on if the target is not found."
    )
    starting_dir: str = Field(
        '.',
        description="The starting directory to search from. Defaults to current working directory.",
    )

    _args: list = ['target']

    def execute(self):
        files = []
        for (dirpath, dirnames, filenames) in os.walk(self.starting_dir):
            files += [
                os.path.join(dirpath, file) for file in filenames if file == self.target
            ]

        if len(files) == 0:
            return self.fallback

        return files


class PathJoinHook(BaseHook):
    """Hook joining paths."""

    hook_type: str = 'path_join'
    paths: list = Field(
        ...,
        description="List of items in a path to file or directory.",
        render_by_default=True,
    )

    _args: list = ['paths']

    def execute(self):
        return os.path.join(*self.paths)



"""High level tests for parser logic."""
import pytest
import yaml

from tackle.main import tackle


FIXTURES = [
    ('map.yaml', 'map-output.yaml'),
    ('map-lists.yaml', 'map-lists-output.yaml'),
    ('map-root.yaml', 'map-root-output.yaml'),
    ('outer-tackle.yaml', 'outer-tackle-output.yaml'),
    ('private-hooks.yaml', 'private-hooks-output.yaml'),
    # # ('document-hooks.yaml', 'outer_tackle_expected.yaml'),
    ('outer-tackle.yaml', 'outer-tackle-output.yaml'),
    ('merge-simple.yaml', 'merge-simple-output.yaml'),
    ('merge-petstore.yaml', 'petstore.yaml'),
    # # Non tackle things
    ('k8s-deployment.yaml', 'k8s-deployment.yaml'),
    ('docker-compose.yml', 'docker-compose.yml'),
    ('list-list.yaml', 'list-list.yaml'),
]


@pytest.mark.parametrize("fixture,expected_output", FIXTURES)
def test_main_expected_output(change_curdir_fixtures, fixture, expected_output):
    """Input equals output."""
    with open(expected_output) as f:
        expected_output = yaml.safe_load(f)

    output = tackle(fixture)
    assert dict(output) == expected_output



"""All exceptions used in the tackle box code base are defined here."""
import inspect
import os


class ContributionNeededException(Exception):
    """
    Special exception to point users (particularly of the windows variety) to contribute
    a fix with context that includes a link to the file needing upgrade.
    """

    def __init__(self, extra_message=None):
        # self.path_to_code = link_to_code
        self.extra_message = extra_message
        self.message = (
            f"Unimplemented / needs to be built "
            f"- PLEASE HELP -> "
            f"https://github.com/robcxyz/tackle-box/blob/main/{self.file_location()}"
            + f"\n{self.extra_message}"
        )
        super().__init__(self.message)

    def __new__(cls, *args, **kwargs):
        new_instance = super(ContributionNeededException, cls).__new__(
            cls, *args, **kwargs
        )
        stack_trace = inspect.stack()
        created_at = '%s:%d' % (stack_trace[1][1], stack_trace[1][2])
        new_instance.created_at = created_at
        return new_instance

    def file_location(self):
        path_list = os.path.normpath(self.created_at).split(os.sep)
        for i, v in enumerate(path_list):
            if v == 'tackle':
                break
        github_path_list = path_list[i:-1] + [
            path_list[-1].split(':')[0] + "#L" + path_list[-1].split(':')[1]
        ]
        return os.path.join(*github_path_list)


class TackleException(Exception):
    """
    Base exception class.

    All Tackle-specific exceptions should subclass this class.
    """

    # def __init__(self, message=None):
    #     # self.path_to_code = link_to_code
    #     self.message = message
    #     super().__init__(self.message)


class EmptyBlockException(TackleException):
    """
    Exception when a block (ie a compact hook without a hook type - key->: if:/for:)
    is declared without any contents.

    Raised when handling empty blocks.
    """


class NonTemplatedInputDirException(TackleException):
    """
    Exception for when a project's input dir is not templated.

    The name of the input directory should always contain a string that is
    rendered to something else, so that input_dir != output_dir.
    """


class UnknownSourceException(TackleException):
    """
    Exception for ambiguous source.

    Raised when tackle cannot determine which directory is the project
    template, e.g. more than one dir appears to be a template dir.
    """

    # unused locally


class MissingProjectDir(TackleException):
    """
    Exception for missing generated project directory.

    Raised during cleanup when remove_repo() can't find a generated project
    directory inside of a repo.
    """

    # unused locally


class UnknownArgumentException(TackleException):
    """
    Exception unknown argument when supplied via a hook call.

    Raised when a hook is called with an unknown argument in it's mapping (ie missing
    `_args` param in the hook definition.
    """


class NoInputOrParentTackleException(TackleException):
    """
    Exception when no input has been given, nor is there a parent tackle.

    Raised when a hook is called with an unknown argument in it's mapping (ie missing
    `_args` param in the hook definition.
    """


class UnknownTemplateVariableException(TackleException):
    """
    Exception an unknown templatable argument.

    Raised when rendering variables.
    """


class EmptyTackleFileException(TackleException):
    """
    Exception when base tackle file is empty.

    Raised when calling files / providers.
    """


class ConfigDoesNotExistException(TackleException):
    """
    Exception for missing config file.

    Raised when get_config() is passed a path to a config file, but no file
    is found at that path.
    """


class InvalidConfiguration(TackleException):
    """
    Exception for invalid configuration file.

    Raised if the global configuration file is not valid YAML or is
    badly constructed.
    """


class UnknownRepoType(TackleException):
    """
    Exception for unknown repo types.

    Raised if a repo's type cannot be determined.
    """


class VCSNotInstalled(TackleException):
    """
    Exception when version control is unavailable.

    Raised if the version control system (git or hg) is not installed.
    """


class VersionNotFoundError(TackleException):
    """
    Exception when a version to a provider is given but does not exist.

    Raised if a version is specified but does not exist on the remote.
    """


class UnsupportedBaseFileTypeException(TackleException):
    """
    Exception for when a none json / yaml file are read

    Raised if the base file that is being called is not json / yaml.
    """


class ContextDecodingException(TackleException):
    """
    Exception for failed JSON decoding.

    Raised when a project's JSON context file can not be decoded.
    """


class OutputDirExistsException(TackleException):
    """
    Exception for existing output directory.

    Raised when the output directory of the project exists already.
    """


class InvalidModeException(TackleException):
    """
    Exception for incompatible modes.

    Raised when tackle is called with both `no_input==True` and
    `replay==True` at the same time.
    """


class FailedHookException(TackleException):
    """
    Exception for hook failures.

    Raised when a hook script fails.
    """


class UndefinedVariableInTemplate(TackleException):
    """
    Exception for out-of-scope variables.

    Raised when a template uses a variable which is not defined in the
    context.
    """

    def __init__(self, message, error, context):
        """Exception for out-of-scope variables."""
        self.message = message
        self.error = error
        self.context = context

    def __str__(self):
        """Text representation of UndefinedVariableInTemplate."""
        return self.message


class UnknownExtension(TackleException):
    """
    Exception for un-importable extention.

    Raised when an environment is unable to import a required extension.
    """


class RepositoryNotFound(TackleException):
    """
    Exception for missing repo.

    Raised when the specified tackle repository doesn't exist.
    """


class ContextFileNotFound(TackleException):
    """
    Exception for missing context file.

    Raised when the specified context file doesn't exist.
    """


class RepositoryCloneFailed(TackleException):
    """
    Exception for un-cloneable repo.

    Raised when a tackle template can't be cloned.
    """


class InvalidZipRepository(TackleException):
    """
    Exception for bad zip repo.

    Raised when the specified tackle repository isn't a valid
    Zip archive.
    """


class InvalidOperatorType(TackleException):
    """
    Exception for bad zip repo.

    Raised when the specified tackle repository isn't a valid
    Zip archive.
    """


class EscapeOperatorException(TackleException):
    """
    Exception for bad zip repo.

    Raised when the specified tackle repository isn't a valid
    Zip archive.
    """


class HookCallException(TackleException):
    """
    Exception for when a unknown field in a hook.

    Raised when field has been provided not declared in the hook type.
    """


class UnknownHookTypeException(TackleException):
    """
    Exception for unknown hook type.

    Raised when a hook type is not available from the providers.
    """


class Exception(TackleException):
    """
    Exception for unknown hook type.

    Raised when a hook type is not available from the providers.
    """


class EscapeHookException(TackleException):
    """
    Exception for when general hook errors.

    Raised in hooks.
    """



"""Tests dict input objects for `tackle.providers.rich.hooks.print` module."""
from tackle.main import tackle


def test_provider_system_hook_print(change_dir):
    """Verify the hook call works properly."""
    output = tackle('print.yaml')
    assert output['md']
    assert output['this'] == 'this and that'
    assert output['multiline'] == 'this and that'



"""Select hook."""
import logging
from PyInquirer import prompt
from typing import Any, List, Union
from pydantic import Field

from tackle.models import BaseHook
from tackle.utils.dicts import get_readable_key_path

logger = logging.getLogger(__name__)


class InquirerListHook(BaseHook):
    """
    Hook for PyInquirer 'list' type prompts.

    Takes in two forms of `choices` inputs.
    1. list of string
    2. list of maps with the key as the output, the value as displayed question

    :param message: String message to show when prompting.
    :param choices: A list of strings or list of k/v pairs per above description
    :param name: A key to insert the output value to. If not provided defaults to
        inserting into parent key
    :param index:
    :return: String for answer
    """

    hook_type: str = 'select'

    message: str = Field(None, description="String message to show when prompting.")
    default: Any = Field(None, description="Default choice.")
    name: str = Field('tmp', description="Extra key to embed into. Artifact of API.")
    choices: Union[List[str], List[dict]] = Field(
        ...,
        description="List of strings or dicts with keys as output and values as display.",
    )
    index: bool = Field(
        False, description="Boolean to return the index instead of the answer"
    )

    _args: list = ['message']

    def __init__(self, **data: Any):
        super().__init__(**data)
        if self.message is None:
            self.message = get_readable_key_path(self.key_path) + ' >>>'

    def execute(self) -> str:
        # Figure out what type of dictionary it is
        choices_type = None
        for i, v in enumerate(self.choices):
            if i != 0 and type(v) != choices_type:
                raise ValueError("All items need to be of the same type")
            choices_type = type(v)

        if choices_type == str:
            answer = self._run_prompt()
            if self.index:
                return self.choices.index(answer)
            else:
                return answer
        elif choices_type == dict:
            choices = []
            for i in self.choices:
                choices.append(i[list(i.keys())[0]])

            choices_dict = self.choices
            self.choices = choices

            answer = self._run_prompt()
            for i in choices_dict:
                val = list(i.keys())[0]
                if i[val] == answer:
                    if self.index:
                        return self.choices.index(answer)
                    else:
                        return val
        else:
            raise ValueError("Choices must be dict with ")

    def _run_prompt(self):
        if not self.no_input:
            question = {
                'type': 'list',
                'name': self.name,
                'message': self.message,
                'choices': self.choices,
            }
            if self.default:
                question.update({'default': self.default})
            response = prompt([question])

            if self.name != 'tmp':
                return response
            else:
                return response['tmp']
        elif self.default:
            return self.default
        elif isinstance(self.choices[0], str):
            return self.choices[0]
        elif isinstance(self.choices[0], dict):
            return list(self.choices[0].keys())[0]



"""Variable hook."""
from typing import Any
from pydantic import Field
from tackle.models import BaseHook


class VarHook(BaseHook):
    """Hook for rendering a variable based on an input. Only useful for rendering as
    otherwise you wouldn't need this hook at all.

    Alternatively just use:

    ```yaml
    key->: {{some_other_key}}
    ```

    Which under the hood is just rewriting the input before processing as this:

    ```yaml
    key->: var "{{some_other_key}}"
    ```

    For unrendered values you would omit any hook logic.
    """

    hook_type: str = 'var'
    input: Any = Field(..., description="Any variable input.", render_by_default=True)
    _args: list = ['input']

    def execute(self):
        return self.input


class TypeHook(BaseHook):
    """Hook for getting the type of a variable."""

    hook_type: str = 'type'
    input: Any = Field(..., description="Any variable input.", render_by_default=True)
    _args: list = ['input']

    def execute(self):
        return type(self.input)



"""Tests dict input objects for `tackle.providers.tackle.block` module."""
from tackle.main import tackle


def test_provider_system_hook_block_tackle(change_dir):
    """Simple block test."""
    output = tackle('basic.yaml', no_input=True)

    assert output['stuff'] == 'here'
    assert 'things' not in output


def test_provider_system_hook_block_embedded_blocks(change_dir):
    """Embedded with multiple blocks."""
    output = tackle('embedded_blocks.yaml', no_input=True)

    assert output['things'] == 'things'


def test_provider_system_hook_block_looped(change_dir):
    """With a for loop."""
    output = tackle('looped.yaml', no_input=True)

    assert len(output['blocker']) == 2


def test_provider_system_hook_block_block_merge(change_dir):
    """Block with a merge."""
    output = tackle('block_merge.yaml', no_input=True)

    assert output['things'] == 'here'
    assert output['foo'] == 'bar'


def test_provider_system_hook_block_block(change_dir):
    """Complex block."""
    output = tackle('block.yaml', no_input=True)

    assert output['block']['things'] == 'here'
    assert output['block']['foo'] == 'bar'



"""pytest fixtures which are globally available throughout the suite."""
import pytest
import os
import shutil

from tackle.settings import settings


@pytest.fixture()
def tmp_move_tackle_dir():
    """Fixture to temporarily move tackle dir where providers are stored."""
    if os.path.isdir(settings.tackle_dir):
        shutil.move(settings.tackle_dir, settings.tackle_dir + '.tmp')
    yield
    shutil.move(settings.tackle_dir + '.tmp', settings.tackle_dir)



"""Parser for importing providers into the runtime."""
import os
import logging
from pathlib import Path
import subprocess
import sys
import importlib.machinery
from pydantic import BaseModel

from tackle.utils.paths import listdir_absolute, work_in
from tackle.settings import settings

from typing import List

logger = logging.getLogger(__name__)

file_path = os.path.dirname(os.path.abspath(__file__))

native_providers = [
    os.path.join(file_path, f)
    for f in os.listdir(file_path)
    if os.path.isdir(os.path.join(file_path, f)) and f != '__pycache__'
]


class Provider(BaseModel):
    """Base provider."""

    path: str = None
    hooks_path: str = None
    hook_types: list = []
    hook_modules: list = []

    name: str = None  # defaults to os.path.basename(path)

    src: str = None
    version: str = None
    requirements: list = []

    def __init__(self, **data):
        super().__init__(**data)


class ProviderList(BaseModel):
    """
    Provider list object that on instantiation imports native (shipping wit tackle box),
    settings (declared in ~/.tackle/settings.yaml), and local (in a local hooks
    directory).
    """

    __root__: List[Provider] = []
    # _functions: dict = None

    @property
    def hook_types(self):
        hook_types = []
        for i in self:
            hook_types.append(i.name)
        return hook_types

    # @property
    # def functions(self):
    #     functions = {}
    #     for i in self:
    #         functions.append(i.functions)
    #     return functions

    def append(self, value: Provider):
        """Override."""
        self.__root__.append(value)

    def __iter__(self):
        """Override."""
        return iter(self.__root__)

    def __getitem__(self, item):
        """Override."""
        return self.__root__[item]

    def __init__(self, **data):
        super().__init__(**data)
        """Import natve, settings, and local providers."""
        # Import native providers in tackle.providers
        self.import_paths(native_providers)
        # Import any providers in settings
        if settings.extra_providers:
            self.import_paths(settings.extra_providers)
        # If there is a hooks directory in the same dir as execution, import those
        if 'hooks' in os.listdir():
            curent_dir_provider_name = Path().parent.absolute().name
            with work_in(".."):
                self.import_paths([curent_dir_provider_name])

    def import_paths(self, provider_paths):
        """Iterate through paths and import them."""
        for i in provider_paths:
            self.append_from_path(i)

    def append_from_path(self, provider_path):
        """Append a provider with a given path."""
        mod_name = 'tackle.providers.' + os.path.basename(provider_path)
        hooks_init_path = os.path.join(provider_path, 'hooks', '__init__.py')
        hooks_path = os.path.join(provider_path, 'hooks')

        # If the provider has an __init__.py in the hooks directory, import that
        # to check if there are any hook types declared there.  If there are, store
        # those references so that if the hook type is later called, the hook can
        # then be imported.
        if os.path.isfile(hooks_init_path):
            # mod = import_module_from_path(mod_name, hooks_init_path)
            loader = importlib.machinery.SourceFileLoader(mod_name, hooks_init_path)
            mod = loader.load_module()

            try:
                hook_types = mod.hook_types
                p = Provider(path=hooks_path, hook_types=hook_types, name=mod_name)
                logger.debug(
                    f"Importing hook_types {mod.hook_types} from {hooks_path}."
                )
                self.append(p)
            except AttributeError:
                logger.debug(
                    f"No `hook_types` declared in {hooks_path}, importing module."
                )
                import_with_fallback_install(mod_name, hooks_path)
        else:
            # This pass will import all the modules and when retrieving hook types,
            # wil search based on all subclasses
            # self.append(get_provider_from_dir(mod_name, hooks_path))
            import_with_fallback_install(mod_name, hooks_path)

            p = Provider(path=hooks_path, name=mod_name)
            self.append(p)


def import_with_fallback_install(mod_name, path):
    """Import a module and on import error, fallback on requirements file."""
    try:
        import_hooks_from_dir(mod_name, path)
    except ModuleNotFoundError:
        requirements_path = os.path.join(path, '..', 'requirements.txt')
        if os.path.isfile(requirements_path):
            # It is a convention of providers to have a requirements file at the base.
            # Install the contents if there was an import error
            subprocess.check_call(
                [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "--quiet",
                    "--disable-pip-version-check",
                    "-r",
                    requirements_path,
                ]
            )
        import_hooks_from_dir(mod_name, path)


# def get_provider_from_dir(mod_name, path):
#     """Initialize the provider from a path."""
#     import_with_fallback_install(mod_name, path)
#     p = Provider(path=path, name=mod_name)
#     return p


def import_hooks_from_dir(
    mod_name,
    path,
    excluded_file_names=None,
    excluded_file_extensions=None,
):
    """
    Import hooks from a directory.

    This is meant to be used by generically pointing to a hooks directory and
    importing all the relevant hooks into the context.
    """
    if excluded_file_names is None:
        # Maintaining cookiecutter support here?
        excluded_file_names = ['pre_gen_project', 'post_gen_project', '__pycache__']
    if excluded_file_extensions is None:
        excluded_file_extensions = ['pyc']

    potential_hooks = listdir_absolute(path)
    for f in potential_hooks:
        file_base = os.path.basename(f).split('.')
        if file_base[0] in excluded_file_names:
            continue
        if file_base[-1] in excluded_file_extensions:
            continue

        if os.path.basename(f).split('.')[-1] != "py":
            # Only import python files
            continue

        loader = importlib.machinery.SourceFileLoader(
            mod_name + '.' + os.path.basename(f).split('.')[0], f
        )
        loader.load_module()



"""Assert hook."""
from tackle.models import BaseHook

from typing import Any


class AssertHook(BaseHook):
    """Hook for asserting an input is equal to a value."""

    hook_type: str = 'assert'
    input: Any
    value: Any

    _args = ['input', 'value']

    def execute(self):
        assert self.input == self.value



"""Tests dict input objects for `tackle.providers.pyinquirer.hooks.confirm` module."""
from tackle.main import tackle


def test_provider_pyinquirer_confirm_hook(change_dir):
    """Verify the hook call works successfully."""
    o = tackle('tackle.yaml', no_input=True)
    assert 'this' in o



"""Print hooks."""
from pprint import pprint
from typing import Any

from tackle import BaseHook, Field


class PrintHook(BaseHook):
    r"""
    Hook  for printing an input and returning the output.

    Follows: https://docs.python.org/3/library/functions.html#print

    print(*objects, sep=' ', end='\n', file=sys.stdout, flush=False)
    """

    hook_type: str = 'print'

    objects: Any = Field(None, description="The objects to print.")
    sep: str = Field(' ', description="Separator between printed objects.")
    end: str = Field('\n', description="What to print at the end")
    flush: bool = Field(False, description="No clue.")

    _args: list = ['objects']

    def execute(self):
        print(self.objects, sep=self.sep, end=self.end, flush=self.flush)
        return self.objects


class PprintHook(BaseHook):
    """
    Wraps python pprint builtin.

    https://docs.python.org/3/library/pprint.html#pprint.PrettyPrinter
    """

    hook_type: str = 'pprint'

    objects: Any = None
    indent: int = 1
    width: int = 80
    depth: int = None
    compact: bool = False
    sort_dicts: bool = True
    underscore_numbers: bool = False

    _args: list = ['objects']

    def execute(self):
        pprint(
            self.objects,
            indent=self.indent,
            width=self.width,
            # stream=self.stream,
            compact=self.compact,
            sort_dicts=self.sort_dicts,
            underscore_numbers=self.underscore_numbers,
        )
        return



"""Tests dict input objects for `tackle.providers.system.hooks.dicts` module."""
import os
from tackle.main import tackle
import pytest


@pytest.fixture()
def clean_outputs():
    """Clean the output files."""
    yield
    files = [f for f in os.listdir() if f.split('.')[0].startswith('output')]
    for f in files:
        os.remove(f)


def test_provider_system_hook_dicts_merge(change_dir):
    """Verify the hook call works properly."""
    output = tackle('merge.yaml')
    assert output['merge_map']['stuff'] == 'bing'


def test_provider_system_hook_dicts_update(change_dir):
    """Verify the hook call works properly."""
    output = tackle('update.yaml')
    assert output['update_map'] == output['arg']
    assert output['update_map2'] == output['arg2']


def test_provider_system_hook_dicts_pop(change_dir, clean_outputs):
    """Verify the hook call works properly."""
    output = tackle('pop.yaml')
    assert 'stuff' not in output['pop_map']
    assert 'things' not in output['pop_maps']
    assert output['arg_1'] == ['stuff']
    assert 'foo' not in output['arg_2']
    assert 'baz' in output['arg_2']


def test_provider_system_hook_dicts_keys(change_dir, clean_outputs):
    """Verify the hook call works properly."""
    output = tackle('keys.yaml')
    assert output['arg_1'] == ['stuff', 'things']



"""Raw list hook."""
from PyInquirer import prompt
from pydantic import Field

from typing import Any
from tackle.models import BaseHook
from tackle.utils.dicts import get_readable_key_path


class InquirerRawListHook(BaseHook):
    """
    Hook for PyInquirer 'rawlist' type prompts.

    Example: https://github.com/CITGuru/PyInquirer/blob/master/examples/rawlist.py
    :param message: String message to show when prompting.
    :param choices: A list of strings or list of k/v pairs per above description
    :param name: A key to insert the output value to. If not provided defaults to
        inserting into parent key
    :return: String for the answer
    """

    hook_type: str = 'rawlist'

    default: Any = Field(None, description="Default choice.")
    message: str = Field(None, description="String message to show when prompting.")
    name: str = Field('tmp', description="Extra key to embed into. Artifact of API.")

    _args: list = ['message', 'default']

    def __init__(self, **data: Any):
        super().__init__(**data)
        if self.message is None:
            self.message = get_readable_key_path(self.key_path) + ' >>>'

    def execute(self):
        if not self.no_input:
            question = {
                'type': self.type,
                'name': self.name,
                'message': self.message,
                'choices': self.choices,
            }
            if self.default:
                question.update({'default': self.default})

            response = prompt([question])
            if self.name != 'tmp':
                return response
            else:
                return response['tmp']
        elif self.default:
            return self.default
        else:
            # When no_input then return empty list
            return []



"""List directory hook."""
from typing import Union, List
import os
import logging

from tackle.models import BaseHook, Field

logger = logging.getLogger(__name__)


class ListdirHook(BaseHook):
    """
    Hook  for `listdir`. Lists the contents of a directory.

    :return: A list of contents of the `path` if input is string,
        A map with keys of items if input `path` is list.
    """

    hook_type: str = 'listdir'
    ignore_hidden_files: bool = Field(
        None, description="Boolean to ignore hidden files"
    )
    path: Union[List[str], str] = Field(
        None, description="String or list to directories to list"
    )
    sort: bool = Field(False, description="Boolean to sort the output")

    _args: list = ['path']

    def execute(self):
        if isinstance(self.path, str):
            files = os.listdir(os.path.expanduser(self.path))
            if self.sort:
                files.sort()
            if self.ignore_hidden_files:
                return [f for f in files if not f.startswith('.')]
            else:
                return files

        if isinstance(self.path, list):
            contents = {}
            for i in self.path:
                contents[i] = os.listdir(os.path.expanduser(i))
                if self.sort:
                    contents[i].sort()
            return contents



"""Test hook model."""
from tackle.models import BaseHook
import logging

LOGGER = logging.getLogger(__name__)


def test_base_hook_init():
    """Test instantiating a hook."""
    inp = {
        'hook_type': 'foo',
        # 'foo': 'bar',
        'else': 'foo',
        'merge': True,
        'if': 'stuff == things',
    }

    h = BaseHook(**inp)
    # assert h.hook_type == 'foo'
    assert h.if_ == '{{stuff == things}}'
    assert h.merge
    assert isinstance(h.merge, bool)


# def test_hook_dict_match_case(caplog):
#     inp = {
#         'match': 'foo',
#         'case': ['bar', '_'],
#     }
#     h = BaseHook(**inp)
#     if sys.version_info[1] >= 10:
#         assert h.match_ == 'foo'
#
#     if sys.version_info[1] <= 10:
#         assert h.match_ is None
#
#     # TODO: Fix this test
# log_output = caplog.text
# assert 'Must be using Python' in caplog.text



"""Tackle hooks."""
import tackle as tkl
from pydantic import SecretStr

from tackle.models import BaseHook, Field


class TackleHook(BaseHook):
    """
    Hook  for calling external tackle.

    :param template: A directory containing a project template,
        or a URL to a git repository.
    :param templates: A list of directories containing a project template,
        or a URL to a git repository.
    :param checkout: The branch, tag or commit ID to checkout after clone.
    :param no_input: Prompt the user at command line for manual configuration?
    :param override_inputs: A dictionary of context that overrides default
        and user configuration.
    :param existing_context: An additional dictionary to use in rendering
        additional prompts.
    :param replay: Do not prompt for input, instead read from saved json. If
        ``True`` read from the ``replay_dir``.
        if it exists
    :param output_dir: Where to output the generated project dir into.
    :param config_file: User configuration file path.
    :param default_config: Use default values rather than a config file.
    :param password: The password to use when extracting the repository.
    :param directory: Relative path to a tackle box / cookiecutter template
        in a repository.
    :param accept_hooks: Accept pre and post hooks if set to `True`.

    :return: Dictionary of output
    """

    hook_type: str = 'tackle'

    # fmt: off

    input_string: str = Field(
        None, description="The input can be one of repo, file path, directory with tackle.yaml, zip file, or if left blank parent tackle file.")
    checkout: str = Field(None, description="The branch or version to checkout for repo type inputs_strings.")
    context_file: str = Field(None, description="The file to run inside a repo input.")
    additional_context: dict = Field(
        None, description="Any additional context to use when calling the hook. Like existing context.")
    context: dict = Field(None, description="A context to use that overrides the current context.")
    password: SecretStr = Field(None, description="A password to use for repo inputs.")
    directory: str = Field(None, description="The directory to run inside for repo inputs.")

    override: dict = Field(None, description="A dictionary of keys to override.")
    # fmt: on

    _args = ['input_string']

    def __init__(self, **data):
        super().__init__(**data)
        if self.override is None:
            self.override = {}

    def execute(self):

        # if self.existing_context in (None, {}):
        #     existing_context = self.output_dict
        # else:
        #     existing_context = self.existing_context

        if self.context:
            existing_context = self.context
        else:
            existing_context = self.output_dict.copy()
            existing_context.update(self.existing_context)

            if self.additional_context:
                existing_context.update(self.additional_context)

        output_context = tkl.main.tackle(
            self.input_string,
            checkout=self.checkout,
            password=self.password,
            directory=self.directory,
            calling_directory=self.calling_directory,
            # Evaluated
            existing_context=existing_context,
            # Implicit
            providers=self.providers,
            no_input=self.no_input,
            global_kwargs=self.override,
        )

        return dict(output_context)



"""Tests dict input objects for `tackle.providers.system.hooks.var` module."""
from tackle.main import tackle


def test_provider_system_hook_split(change_dir):
    """Verify the hook call works properly."""
    output = tackle('tackle.yaml')

    assert output['a_str'] == ['things', 'stuffs']
    assert output['join_a_str'] == 'things.stuffs'



# """Collection of tests around repository type identification."""
# import pytest
#
# from tackle import exceptions
# from tackle.utils import vcs
#
#
# @pytest.mark.parametrize(
#     'repo_url, exp_repo_type, exp_repo_url',
#     [
#         (
#             'git+https://github.com/pytest-dev/cookiecutter-pytest-plugin.git',
#             'git',
#             'https://github.com/pytest-dev/cookiecutter-pytest-plugin.git',
#         ),
#         (
#             'hg+https://bitbucket.org/foo/bar.hg',
#             'hg',
#             'https://bitbucket.org/foo/bar.hg',
#         ),
#         (
#             'https://github.com/pytest-dev/cookiecutter-pytest-plugin.git',
#             'git',
#             'https://github.com/pytest-dev/cookiecutter-pytest-plugin.git',
#         ),
#         ('https://bitbucket.org/foo/bar.hg', 'hg', 'https://bitbucket.org/foo/bar.hg'),
#         (
#             'https://github.com/audreyr/cookiecutter-pypackage.git',
#             'git',
#             'https://github.com/audreyr/cookiecutter-pypackage.git',
#         ),
#         (
#             'https://github.com/audreyr/cookiecutter-pypackage',
#             'git',
#             'https://github.com/audreyr/cookiecutter-pypackage',
#         ),
#         (
#             'git@gitorious.org:cookiecutter-gitorious/cookiecutter-gitorious.git',
#             'git',
#             'git@gitorious.org:cookiecutter-gitorious/cookiecutter-gitorious.git',
#         ),
#         (
#             'https://audreyr@bitbucket.org/audreyr/cookiecutter-bitbucket',
#             'hg',
#             'https://audreyr@bitbucket.org/audreyr/cookiecutter-bitbucket',
#         ),
#     ],
# )
# def test_identify_known_repo(repo_url, exp_repo_type, exp_repo_url):
#     """Verify different correct repositories url syntax is correctly transformed."""
#     assert vcs.identify_repo(repo_url) == (exp_repo_type, exp_repo_url)
#
#
# @pytest.fixture(
#     params=[
#         'foo+git',  # uses explicit identifier with 'git' in the wrong place
#         'foo+hg',  # uses explicit identifier with 'hg' in the wrong place
#         'foo+bar',  # uses explicit identifier with neither 'git' nor 'hg'
#         'foobar',  # no identifier but neither 'git' nor 'bitbucket' in url
#         'http://norepotypespecified.com',
#     ]
# )
# def unknown_repo_type_url(request):
#     """Fixture. Return wrong formatted repository url."""
#     return request.param
#
#
# def test_identify_raise_on_unknown_repo(unknown_repo_type_url):
#     """Verify different incorrect repositories url syntax trigger error raising."""
#     with pytest.raises(exceptions.UnknownRepoType):
#         vcs.identify_repo(unknown_repo_type_url)



# """Collection of tests around VCS detection."""
# import pytest
#
# from tackle.utils import vcs
#
#
# @pytest.mark.parametrize(
#     'which_return, result',
#     [('', False), (None, False), (False, False), ('/usr/local/bin/git', True)],
# )
# def test_is_vcs_installed(mocker, which_return, result):
#     """Verify `is_vcs_installed` function correctly handles `which` answer."""
#     mocker.patch('tackle.utils.vcs.which', autospec=True, return_value=which_return)
#     assert vcs.is_vcs_installed('git') == result



"""Tests dict input objects for `tackle.providers.pyinquirer.hooks.checkbox` module."""
from tackle.main import tackle


# TODO: Need to properly test this with pty. Tests don't cover now
def test_provider_checkbox_dict_ok(change_dir):
    """Verify the hook call works successfully."""
    context_dict_ok = tackle('dict_ok.yaml', no_input=True)
    assert context_dict_ok


def test_provider_checkbox_dict_index(change_dir):
    """Verify the hook call works successfully."""
    context_string_ok = tackle('dict_index.yaml', no_input=True)

    assert context_string_ok


def test_provider_checkbox_string_ok(change_dir):
    """Verify the hook call works successfully."""
    context_dict_ok = tackle('string_ok.yaml', no_input=True)

    assert context_dict_ok



"""Command hook."""
import sys
import logging
import subprocess
import errno
import struct
import shutil
import os
from itertools import chain
from select import select
from pydantic import Field

from tackle.models import BaseHook
from tackle.exceptions import HookCallException

if os.name != 'nt':
    # Don't import on windows as pty is not available there
    import pty
    import fcntl
    import termios


logger = logging.getLogger(__name__)


class CommandHook(BaseHook):
    """System calls."""

    hook_type: str = 'command'

    command: str = Field(..., description="A shell command.")
    ignore_error: bool = Field(False, description="Ignore errors.")

    _args: list = ['command']

    def execute(self):
        p = subprocess.Popen(
            self.command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        output, err = p.communicate()

        if err and not self.ignore_error:
            raise HookCallException(err.decode('utf-8'))
        if err and self.ignore_error:
            return err.decode('utf-8')

        return output.decode("utf-8")


class ShellHook(BaseHook):
    """
    `shell` hook for system calls.

    Streams the output of the process.

    :param command: The command to run on the host
    :return: String output of command
    """

    hook_type: str = 'shell'

    command: str
    ignore_error: bool = False

    _args: list = ['command']

    def _set_size(self, fd):
        """Found at: https://stackoverflow.com/a/6420070."""
        (
            _COLUMNS,
            _ROWS,
        ) = shutil.get_terminal_size(fallback=(80, 20))
        size = struct.pack("HHHH", _ROWS, _COLUMNS, 0, 0)
        fcntl.ioctl(fd, termios.TIOCSWINSZ, size)

    def execute(self):
        # Copied from
        # https://terminallabs.com/blog/a-better-cli-passthrough-in-python/
        masters, slaves = zip(pty.openpty(), pty.openpty())
        for fd in chain(masters, slaves):
            self._set_size(fd)

        # cmd = re.findall(r'\S+|\n', self.command)
        # cmds = self.command.split('\n')
        # for cmd in cmds:
        # TODO: Fix multi-line calls

        cmd = self.command.replace('\n', ' ; ').split()

        try:
            with subprocess.Popen(
                cmd, stdin=slaves[0], stdout=slaves[0], stderr=slaves[1]
            ) as p:
                for fd in slaves:
                    os.close(fd)  # no input
                    readable = {
                        masters[0]: sys.stdout.buffer,  # store buffers seperately
                        masters[1]: sys.stderr.buffer,
                    }
                while readable:
                    for fd in select(readable, [], [])[0]:
                        try:
                            data = os.read(fd, 1024)  # read available
                        except OSError as e:
                            if e.errno != errno.EIO:
                                raise  # XXX cleanup
                            del readable[fd]  # EIO means EOF on some systems
                        else:
                            if not data:  # EOF
                                del readable[fd]
                            else:
                                if fd == masters[0]:  # We caught stdout
                                    print(data.rstrip())
                                else:  # We caught stderr
                                    print(data.rstrip(), err=True)
                                readable[fd].flush()
            for fd in masters:
                os.close(fd)
            logger.debug("Process exited with %s return code." % p.returncode)
            if p.returncode != 0 and not self.ignore_error:
                raise HookCallException(data.decode('utf-8'))
            else:
                return data.decode("utf-8")

        except Exception as e:
            if self.ignore_error:
                return e
            else:
                raise HookCallException(e)



"""Dict hooks."""
import logging
from typing import Union, Dict, List

from tackle import BaseHook, Field
from tackle.utils import merge_configs

logger = logging.getLogger(__name__)


class DictUpdateHook(BaseHook):
    """Hook for updating dict objects with items."""

    hook_type: str = 'update'

    src: dict = Field(description="A dict or list of dicts to update the input `src`")
    input: dict = Field(description="A dict or list of dicts to update the input `src`")

    _args: list = ['src', 'input']

    def execute(self):
        self.src.update(self.input)
        return self.src


class DictMergeHook(BaseHook):
    """
    Hook for recursively merging dict objects with input maps.

    :param src: The input dict to update
    :param input: A dict or list of dicts to update the input `dict`
    :return: An updated dict object.
    """

    hook_type: str = 'merge'
    inputs: list = Field(
        ...,
        description="A list of dictionaries to merge together.",
        render_by_default=True,
    )
    # pairs: Union[dict, list] = Field(..., description="")

    _args: list = ['inputs']
    _render_by_default = ['pairs']

    def execute(self):
        output = {}
        for i in self.inputs:
            output = merge_configs(output, i)
        return output


class DictPopHook(BaseHook):
    """
    Hook  for recursively merging dict objects with input maps.

    :param src: The input dict to update
    :param item: A list or string of items to remove from a dictionary or list
    :return: An updated dict object.
    """

    hook_type: str = 'pop'

    src: Union[dict, list] = None
    item: Union[Dict, List[str], str] = None

    _args: list = ['src', 'item']

    def execute(self):
        if self.item is None:
            self.src.pop()
        elif isinstance(self.item, list):
            for i in self.item:
                self.src.pop(i)
        else:
            self.src.pop(self.item)

        return self.src


class DictKeysHook(BaseHook):
    """
    Hook  for returning the keys of a dict as a list.

    :param src: The input dict or list of dicts return the keys for
    :return: List of keys or list of list of keys if input is list
    """

    hook_type: str = 'keys'

    src: dict = None

    _args: list = ['src']

    def execute(self):
        if isinstance(self.src, list):
            keys = []
            for i in self.src:
                keys.append(i.keys())
            return keys
        else:
            return list(self.src.keys())



"""Tests dict input objects for `tackle.providers.system.hooks.random` module."""
from tackle.main import tackle


def test_provider_system_hook_random_string(change_dir):
    """Verify the hook call works properly."""
    output = tackle()

    assert len(output['random_hex']) == 8
    assert len(output['random_hex_arg']) == 4
    assert len(output['random_string']) == 8
    assert len(output['random_string_arg']) == 6
    assert output['random_string_lower_arg'].isupper()



"""Tests dict input objects for `tackle.providers.system.hooks.command` module."""
import os
import pytest
from tackle.exceptions import HookCallException

from tackle.main import tackle

if os.name == 'nt':
    pytest.skip("Skipping when run from windows.", allow_module_level=True)


def test_provider_system_hook_command(change_dir):
    """Verify the hook call works properly."""
    fixture = 'list-dir.yaml'
    context = tackle(fixture)
    # assert fixture in context['cmd']
    # assert context['shell']
    assert context['cmd'] == context['cmd_arg']


def test_provider_system_hook_command_multi_line(change_dir):
    """Verify the hook call works properly."""
    context = tackle('multi-line-cmd.yaml')
    assert 'No such file' not in context['shell']


def test_provider_system_hook_command_exit(change_dir):
    """Verify the hook call works properly."""
    with pytest.raises(HookCallException):
        tackle('command-exit.yaml')


def test_provider_system_hook_shell_exit(change_dir):
    """Verify the hook call works properly."""
    with pytest.raises(HookCallException):
        tackle('shell-exit.yaml')


def test_provider_system_hook_shell_exit_long(change_dir):
    """Verify the hook call works properly."""
    with pytest.raises(HookCallException):
        tackle('shell-exit-long.yaml')


def test_provider_system_hook_shell_exit_ignore(change_dir):
    """Verify the hook call works properly."""
    o = tackle('shell-exit-ignore.yaml')
    assert o


def test_provider_system_hook_command_exit_ignore(change_dir):
    """Verify the hook call works properly."""
    o = tackle('command-exit-ignore.yaml')
    assert o



"""Collection of tests around loading extensions."""
import pytest

from tackle.render.environment import StrictEnvironment
from tackle.exceptions import UnknownExtension
from tackle.main import tackle


def test_env_should_raise_for_unknown_extension():
    """Test should raise if extension not installed in system."""
    context = {'_extensions': ['foobar']}

    with pytest.raises(UnknownExtension) as err:
        StrictEnvironment(context=context, keep_trailing_newline=True)

    assert 'Unable to load extension: ' in str(err.value)


def test_env_should_come_with_default_extensions():
    """Verify default extensions loaded with StrictEnvironment."""
    env = StrictEnvironment(keep_trailing_newline=True)
    # assert 'jinja2_time.jinja2_time.TimeExtension' in env.extensions
    assert 'tackle.render.extensions.JsonifyExtension' in env.extensions
    assert 'tackle.render.extensions.RandomStringExtension' in env.extensions


def test_env_should_evaluate_is_defined(change_curdir_fixtures):
    """Verify that the `is defined` works when rendering."""
    o = tackle('is_defined.yaml', no_input=True)
    assert 'defined' in o
    assert 'not_defined' not in o
    assert 'not_defined_again' not in o
    assert 'defined_again' in o



"""String hooks."""
import logging
from typing import List

from tackle.models import BaseHook, Field

logger = logging.getLogger(__name__)


class SplitHook(BaseHook):
    """Hook for splitting a string into as list based on a separator."""

    hook_type: str = 'split'

    input: str = Field(..., description="A list of string to split or just a string")
    separator: str = Field(".", description="String separator")
    _args: list = ['input']

    def execute(self):
        return self.input.split(self.separator)


class JoinHook(BaseHook):
    """Join a list of strings with a separator."""

    hook_type: str = 'join'

    input: List[str] = Field(
        ..., description="A list of strings to join.", render_by_default=True
    )
    separator: str = Field('.', description="String separator.")
    _args: list = ['input']

    def execute(self):
        return self.separator.join(self.input)



"""Jinja hook."""
from jinja2.exceptions import UndefinedError
from jinja2 import FileSystemLoader

import logging
from typing import Union

from tackle import BaseHook, Field
from tackle.exceptions import UndefinedVariableInTemplate
from tackle.render.environment import StrictEnvironment

logger = logging.getLogger(__name__)


class JinjaHook(BaseHook):
    """Hook for jinja templates. Returns string path to the output file."""

    hook_type: str = 'jinja'
    file_system_loader: str = Field('.', description="")
    template_path: str = Field(..., description="Path to the template to render")
    output_path: str = Field(..., description="Path to the output file")
    context: Union[dict, str] = Field(None, description="")
    extra_context: dict = Field({}, description="A dict to use to render")
    render_context: dict = Field(
        None, description="A render context that invalidates the default context."
    )
    additional_context: dict = Field(
        None, description="A map to use as additional context when rendering."
    )

    _args: list = ['template_path', 'output_path']

    def __init__(self, **data):
        super().__init__(**data)
        if self.render_context is not None:
            pass
        elif self.additional_context is not None:
            self.render_context = {
                **self.output_dict,
                **self.additional_context,
                **self.existing_context,
            }
        else:
            self.render_context = {
                **self.output_dict,
                **self.existing_context,
            }

    def execute(self) -> dict:
        env = StrictEnvironment(context=self.input_dict)
        env.loader = FileSystemLoader(self.file_system_loader)
        template = env.get_template(self.template_path)

        jinja_context = self.output_dict.copy()

        if self.extra_context:
            jinja_context.update(self.extra_context)

        try:
            # output_from_parsed_template = template.render(
            #     **{self.context_key: jinja_context}, **jinja_context
            # )
            output_from_parsed_template = template.render(**jinja_context)

        except UndefinedError as err:
            msg = f"The Jinja hook for '{self.key}' failed to render"
            raise UndefinedVariableInTemplate(msg, err, self.output_dict)

        with open(self.output_path, 'w') as fh:
            fh.write(output_from_parsed_template)

        return self.output_path



"""Print hooks."""
import logging
from rich.console import Console
from rich.markdown import Markdown

from tackle import BaseHook

logger = logging.getLogger(__name__)


class MarkdownPrintHook(BaseHook):
    """Hook for printing markdown and returning the output."""

    hook_type: str = 'markdown'
    text: str = None

    _args: list = ['text']

    def execute(self):
        console = Console()
        console.print(Markdown(self.text))
        return self.text



"""Match hook."""
from tackle.models import BaseHook, Context, Field
from tackle.parser import walk_sync


class MatchHook(BaseHook):
    """
    Hook for match / case statements. Takes a dict where the keys are matched to a
    value. If the value has an arrow in it (ie key->: ... ) the arrow is stripped away.
    All matched values are ran as hooks.
    """

    hook_type: str = 'match'
    value: str
    case: dict = Field(
        ...,
        description="A dictionary where the keys are cases to be matched. Runs hooks "
        "if present.",
    )

    _args = ['value']

    def execute(self):
        # Condition catches everything except expanded hook calls and blocks (ie key->)
        if self.value in self.case:
            return self.run_key()[self.value]
        else:
            # Iterate through keys and find keys matching value without arrow (->, _>)
            for k, v in self.case.items():
                if self.value == k[:-2]:
                    # Rewrite the value to match the key
                    self.value = k
                    output_dict = self.run_key()
                    # Return the value indexed without arrow
                    return output_dict[k[:-2]]
            raise Exception()
            # raise NoCasesMatchedException()

    def run_key(self):
        case_value = {self.value: self.case[self.value]}

        # Bring in the current input dict
        existing_context = self.output_dict.copy()
        existing_context.update(self.existing_context)

        # Create a temporary context
        tmp_context = Context(
            providers=self.providers,
            existing_context=existing_context,
            output_dict={},
            input_dict=case_value,
            key_path=[],
            no_input=self.no_input,
            calling_directory=self.calling_directory,
        )
        # Traverse the input and update the output dict
        walk_sync(tmp_context, element=case_value.copy())
        # Reindex the return based on self.value
        return tmp_context.output_dict



"""Generate hook."""
import os.path
import fnmatch
from typing import Union, Any
from pydantic import Field
from pathlib import Path
from jinja2 import FileSystemLoader
from jinja2.exceptions import UndefinedError, TemplateNotFound
import shutil

from tackle.models import BaseHook
from tackle.render.environment import StrictEnvironment
from tackle.exceptions import UndefinedVariableInTemplate


class GenerateHook(BaseHook):
    """
    Hook for generating project outputs. Recursively renders all files and folders in a
    given target directory to an output. If there is a "templates" directory and a file
    or directory that matches the `templates` input param, use that as target.
    """

    hook_type: str = 'generate'
    templates: Union[str, list] = Field(
        ...,
        description="Path or list of paths to a templatable directory or file to "
        "recursively render the contents.",
    )
    output: str = Field('.', description="Path to put the output file(s).")
    copy_without_render: Union[str, list] = []
    overwrite_if_exists: bool = Field(
        False, description="Overwrite the output if exists."
    )
    skip_if_file_exists: bool = Field(
        False, description="Skip creating if path exists."
    )
    render_context: dict = Field(
        None, description="A render context that invalidates the default context."
    )
    additional_context: dict = Field(
        None, description="A map to use as additional context when rendering."
    )

    base_dir: Path = None
    env_: Any = None
    file_system_loader_: Any = None
    file_path_separator_: str = None  # / for mac / linux - \ for win

    _args = ['templates', 'output']
    # _render_exclude = ['templates']

    def __init__(self, **data: Any):
        super().__init__(**data)
        if isinstance(self.copy_without_render, str):
            self.copy_without_render = [self.copy_without_render]

        self.output = os.path.expanduser(os.path.expandvars(self.output))
        if 'nt' in os.name:
            self.file_path_separator_ = '\\'
            if not self.output.startswith('\\'):
                self.output = os.path.join(self.calling_directory, self.output)
        else:
            self.file_path_separator_ = '/'
            if not self.output.startswith('/'):
                self.output = os.path.join(self.calling_directory, self.output)

    def execute(self):
        """Generate files / directories."""
        if self.render_context is not None:
            pass
        elif self.additional_context is not None:
            self.render_context = {
                **self.output_dict,
                **self.additional_context,
                **self.existing_context,
            }
        else:
            self.render_context = {
                **self.output_dict,
                **self.existing_context,
            }

        self.env_ = StrictEnvironment(context=self.render_context)
        self.env_.loader = FileSystemLoader('.')

        if isinstance(self.templates, str):
            self.generate_target(self.templates)
        elif isinstance(self.templates, list):
            for target in self.templates:
                self.generate_target(target)

    def generate_target(self, target: str):
        """
        Generate from an unknown target. If there is a `templates` directory in the
        provider, check if the target matches and use that otherwise search for target.
        Generates files or directories from there.

        :param target: A generic target to generate from, file or directory.
        """
        # Update the input to default to searching in `templates` directory
        if os.path.exists(target):
            self.base_dir = Path(".")
        elif os.path.exists(os.path.join("templates", target)):
            self.base_dir = Path("templates")
        else:
            self.base_dir = Path(target).parent.absolute()

        target_path = os.path.join(self.base_dir, target)

        # Expand the target path
        self.output = os.path.expanduser(os.path.expandvars(self.output))
        if self.output == '.':
            self.output = os.path.basename(target)

        if os.path.isfile(target_path):
            self.generate_file(target_path, self.output)

        elif os.path.isdir(target_path):
            self.generate_dir(target_path, self.output)
        else:
            raise TemplateNotFound(f"Could not find {target_path}.")

    def generate_file(self, input_file: str, output_path: str):
        """
        Take a target input_file and render it's contents / file name to an output path.

        :param input_file: Input file to generate from
        :param output_path: Output file to generate to
        """
        if output_path.endswith(self.file_path_separator_):
            output_path = os.path.join(output_path, os.path.basename(input_file))

        # Render the path right away as templating mangles things later - also logical
        # to render file names.  Who wants to generate files with templates in the name?
        file_name_template = self.env_.from_string(str(output_path))
        output_path = file_name_template.render(self.output_dict)

        # Make the parent directories by default
        parent_dir = Path(output_path).parent.absolute()
        Path(parent_dir).mkdir(parents=True, exist_ok=True)

        # Evaluate if it is a copy without render
        if self.is_copy_only_path(input_file):
            shutil.copyfile(input_file, output_path)
            return

        try:
            file_contents_template = self.env_.get_template(input_file)
        except UnicodeDecodeError:
            # Catch binary files with this hack and copy them over
            # TODO: Perhaps improve? In cookiecutter they used a package binary-or-not
            # or something like that but we are staying lean on dependencies in this
            # project.
            shutil.copyfile(input_file, output_path)
            return

        try:
            rendered_contents = file_contents_template.render(self.render_context)
        except UndefinedError as e:
            raise UndefinedVariableInTemplate(
                message=f"Error rendering {input_file}, {e.args[0]}.",
                error=e,
                context=self.render_context,
            )

        # Write contents
        with open(output_path, 'w') as f:
            f.write(rendered_contents)

    def generate_dir(self, input_directory: str, output_path: str):

        for i in os.listdir(input_directory):
            input = os.path.join(input_directory, i)
            output = os.path.join(output_path, i)
            if os.path.isdir(input):
                # Path(output).mkdir(parents=True, exist_ok=True)
                self.generate_dir(input, output)

            if os.path.isfile(input):
                self.generate_file(input, output)

    def is_copy_only_path(self, path):
        """Check whether the given `path` should only be copied and not rendered.

        Returns True if `path` matches a pattern in the given `context` dict,
        otherwise False.

        :param path: A file-system path referring to a file or dir that
            should be rendered or just copied.
        """
        if self.copy_without_render is None:
            return False

        for dont_render in self.copy_without_render:
            if fnmatch.fnmatch(path, dont_render):
                return True
            # TODO: Make this more logical - cookiecutter allowed without `./`
            if fnmatch.fnmatch(path, f'.{self.file_path_separator_}' + dont_render):
                return True
        return False



"""GCP hooks."""
import logging
import os
import importlib.machinery

import json
import inspect
from pydantic import BaseModel, Field
from typing import List, Union, get_type_hints

from tackle.models import BaseHook, Context
from tackle.utils.paths import listdir_absolute

logger = logging.getLogger(__name__)


# Source: https://www.geeksforgeeks.org/python-split-camelcase-string-to-individual-strings/
def camel_case_split(str):
    """Split up camel case string."""
    words = [[str[0]]]

    for c in str[1:]:
        if words[-1][-1].islower() and c.isupper():
            words.append(list(c))
        else:
            words[-1].append(c)

    return [''.join(word) for word in words]


def get_hook_title(hook: BaseHook, schema: dict):
    try:
        return hook.Config.title
    except AttributeError:
        return ' '.join(camel_case_split(schema['title'])[:-1])


class HookDocField(BaseModel):
    name: str
    required: str
    hook_type: str
    default: Union[str, dict, bool, list] = ""
    description: str = ""


class HookDoc(BaseModel):
    hook_type: str
    description: str = ""
    properties: List[HookDocField]
    output: str = None


class ProviderDocs(BaseModel):
    name: str = None
    hooks: List[HookDoc] = []
    requirements: list = []


def get_hook_properties(schema: dict) -> List[HookDocField]:
    from tackle.models import HookDict

    # # h = HookDict()
    # x = BaseHook(type='tmp', hook_dict=HookDict()).schema_json()
    basehook_properties = json.loads(
        BaseHook(type='tmp', hook_dict=HookDict()).schema_json()
    )['properties']

    output = []
    for k, v in schema['properties'].items():
        # Bypass any refs
        if "$ref" in v:
            continue

        # Filter out properties with the same schema as a base property
        # If the field has a description, assume that is meant to be in
        # the docs.
        if k in basehook_properties and 'description' not in v:
            if basehook_properties[k] == v:
                continue

        # Type is dealt with elsewhere
        if k == 'type':
            continue

        if 'anyOf' in v:
            type = ", ".join([i['type'] for i in v['anyOf']])
        else:
            type = v['type']

        hook_doc = HookDocField(
            name=k,
            required="X" if 'default' in v else "",
            type=type,
            default=v['default'] if 'default' in v else "",
            description=v['description'] if 'description' in v else "",
        )
        output.append(hook_doc)

    return output


class ProviderDocsHook(BaseHook):
    """Hook building provider docs."""

    hook_type: str = "docs"

    path: str = Field(".", description="The path to the provider.")
    output: str = Field(".", description="The path to output the docs to.")

    provider: str = Field(None, description="The provider name.")
    hooks_dir = Field("hooks", description="Directory hooks are in.")

    _args: list = ['path', 'output_path']

    def __init__(self, **data):
        super().__init__(**data)
        if not self.provider:
            self.provider = os.path.basename(self.path)

    def execute(self):
        """Build the docs."""
        docs = ProviderDocs()

        # TODO: update the docs with provider metadata
        # Instantiate a context
        # context_fields = [i for i, _ in BaseHook().dict().items()] # + [i for i, _ in BaseHook().dict().items()]

        path = os.path.abspath(os.path.join(self.path, self.hooks_dir))
        # Loop through python files
        for f in [i for i in listdir_absolute(path) if i.endswith("py")]:
            # Extract out the objects that are derived from the BaseHook
            loader = importlib.machinery.SourceFileLoader(self.provider, f)
            module_classes = inspect.getmembers(loader.load_module(), inspect.isclass)
            hooks = [
                i[1]
                for i in module_classes
                if issubclass(i[1], BaseHook) and i[1] != BaseHook
            ]

            for h in hooks:
                # Generate the json schema
                schema = json.loads(h.schema_json())

                # Generate the docs object
                hook_doc = HookDoc(
                    properties=get_hook_properties(schema),
                    type=inspect.signature(h).parameters['type'].default,
                    description=schema['description']
                    if 'description' in schema
                    else "",
                    output=get_type_hints(h.execute)['return'].__name__
                    if 'return' in get_type_hints(h.execute)
                    else None,
                )
                docs.hooks.append(hook_doc)

        return docs.dict()



"""Tests dict input objects for `tackle.providers.system.hooks.lists` module."""
from tackle.main import tackle


def test_provider_system_hook_lists(change_dir):
    """Verify the hook call works properly."""
    output = tackle('tackle.yaml')

    assert 'donkey' in output['appended_list']
    assert 'donkey' in output['appended_lists']
    assert 'chickens' in output['appended_lists']



"""Tests `import` in the `tackle.providers.tackle.hooks.match` hook."""
from tackle import tackle


def test_provider_system_hook_match(change_dir):
    """Run the source and check that the hooks imported the demo module."""
    context = tackle('values.yaml')
    assert 'assert' in context



"""Confirm hook."""
import logging
from PyInquirer import prompt
from typing import Any
from pydantic import Field

from tackle.models import BaseHook
from tackle.utils.dicts import get_readable_key_path

logger = logging.getLogger(__name__)


class InquirerConfirmHook(BaseHook):
    """Hook to confirm with a message and return a boolean."""

    hook_type: str = 'confirm'

    default: Any = Field(True, description="Default choice.")
    message: str = Field(None, description="String message to show when prompting.")
    name: str = Field('tmp', description="Extra key to embed into. Artifact of API.")

    _args: list = ['message']

    def __init__(self, **data: Any):
        super().__init__(**data)
        if self.message is None:
            self.message = get_readable_key_path(self.key_path) + ' >>>'

    def execute(self) -> bool:
        if not self.no_input:
            question = {
                'type': 'confirm',
                'name': self.name,
                'message': self.message,
                'default': self.default,
            }
            response = prompt([question])
            if self.name != 'tmp':
                return response
            else:
                return response['tmp']
        elif self.default:
            return self.default
        else:
            # When no_input then return empty list
            return True



"""Tests dict input objects for `tackle.providers.pyinquirer.hooks.select` module."""

import os
from tackle.main import tackle


def test_provider_select(monkeypatch):
    """Verify the hook call works successfully."""
    monkeypatch.chdir(os.path.abspath(os.path.dirname(__file__)))

    # TODO: Need to properly test this with pty. Tests don't cover now
    output = tackle('dict_ok.yaml', no_input=True)
    assert output

    output = tackle('string_index.yaml', no_input=True)
    assert output

    output = tackle('dict_index.yaml', no_input=True)
    assert output

    output = tackle('list_ok.yaml', no_input=True)
    assert output



"""Hook plugin that inherits a base class and is made available through `type`."""
from __future__ import print_function
from __future__ import unicode_literals

import logging
import re
import toml
from typing import Union, Dict, List, Any

from tackle.models import BaseHook
from tackle.utils import merge_configs

logger = logging.getLogger(__name__)


class TomlHook(BaseHook):
    """
    Hook for toml.

    :param path: The file path to put read or write to
    :param contents: Supplied dictionary or list to write.
    :param in_place: Boolean to read the contents of the `path` and then write after
        modifications.
    :param remove: Parameter or regex to remove from list or dict
    :param update: Use the python `update` dict method on `contents` before writing
    :param filter: List or string to values to.
    :param merge_dict: Dict input that recursively overwrites the `contents`.
    :param append_items: List to append to `append_key` key.
    :param append_key: String or list of hierarchical keys to append item to. Defaults
        to root element.
    :param mode: The mode that the file should write. Defaults to write 'w'.
        Seee https://docs.python.org/3/library/functions.html#open
    """

    hook_type: str = 'toml'

    remove: Union[List, str] = None
    contents: Union[Dict, List] = None
    update: Dict = None
    filter: List = None
    path: str
    merge_dict: Dict = None
    in_place: bool = False
    append_items: Union[Dict, str, List[Any]] = None
    append_keys: Union[Dict, str, List[Any]] = None
    mode: str = None
    write: bool = None

    def execute(self):

        # Load the path into contents unless it already exists
        self._load_contents()
        # Run all the modifiers
        self._modify_dicts()

        if self.write:
            mode = self.mode or 'w'
            with open(self.path, mode) as f:
                toml.dump(self.contents, f)
            return self.contents
        else:
            # Read operation, just return contents
            return self.contents

    def _load_contents(self):
        if self.contents:
            # We are writing. Context is provided
            self.write = True
            pass
        if self.in_place:
            # We are modifying in place. Context is read from path
            self.write = True
            # with open(self.path, 'r') as f:
            #     self.contents = toml.loads(f)
            self.contents = toml.loads(self.path)
        elif not self.contents:
            # We are reading. Contents is read from path
            self.write = False
            # mode = self.mode or 'r'
            # with open(self.path, mode) as f:
            #     self.contents = toml.loads(f)
            self.contents = toml.load(self.path)

    def _modify_dicts(self):
        if self.remove:
            if isinstance(self.remove, str):
                self._remove_from_contents(self.remove)

            if isinstance(self.remove, list):
                for i in self.remove:
                    self._remove_from_contents(i)

        if self.filter:
            if isinstance(self.contents, dict):
                self.contents = {
                    k: v for (k, v) in self.contents.items() if k in self.filter
                }

        if self.update:
            self.contents.update(self.update)

        if self.merge_dict:
            self.contents = merge_configs(self.contents, self.merge_dict)

        if self.append_items:
            if isinstance(self.append_items, str) or isinstance(
                self.append_items, dict
            ):
                self._append_each_item(self.append_items)
            elif isinstance(self.append_items, list):
                for i in self.append_items:
                    self._append_each_item(i)

    def _append_each_item(self, append_item):
        if isinstance(self.append_keys, str):
            self.contents[self.append_keys].append(append_item)
        elif isinstance(self.append_keys, list):
            for k in self.append_keys[:-1]:
                self.contents = self.contents.setdefault(k, {})
            self.contents[self.append_keys[-1]].append(append_item)
        else:
            self.contents.append(append_item)

    def _remove_from_contents(self, regex):
        if isinstance(self.contents, list):
            self.contents = [i for i in self.contents if not re.search(regex, i)]
        if isinstance(self.contents, dict):
            for k in list(self.contents.keys()):
                if re.search(regex, k):
                    self.contents.pop(k)



"""Test hook block rewrite logic in tackle.parser handle_empty_blocks()."""
import pytest
import os
import yaml

from tackle import tackle
from tackle.exceptions import EmptyBlockException

FIXTURES = [
    ('single-level.yaml', 'single-level-output.yaml'),
    ('nested-dict.yaml', 'nested-dict-output.yaml'),
    ('nested-for.yaml', 'nested-for-output.yaml'),
]


@pytest.mark.parametrize("input,output", FIXTURES)
def test_parser_blocks_validate_output(chdir, input, output):
    """Test blocks."""
    chdir(os.path.join("fixtures", "blocks"))
    with open(output) as f:
        expected_output = yaml.safe_load(f)

    tackle_output = tackle(input)
    assert tackle_output == expected_output


ERROR_SOURCES = [
    ("empty-block-exception.yaml", EmptyBlockException),
    # ("")
]


@pytest.mark.parametrize("input_file,exception", ERROR_SOURCES)
def test_parser_raises_exceptions(chdir, input_file, exception):
    """Validate that exceptions are raised."""
    chdir(os.path.join("fixtures", "blocks"))
    with pytest.raises(exception):
        tackle(input_file)



"""Web hooks."""
import logging
import webbrowser
from pydantic import AnyUrl, Field

from tackle.models import BaseHook

logger = logging.getLogger(__name__)


class WebBrowserHook(BaseHook):
    """Hook for registering a variable based on an input. Useful with rendering."""

    hook_type: str = 'webbrowser'
    url: AnyUrl = Field(..., description="String url to open in browser.")

    def execute(self):
        webbrowser.open(self.url, new=2)



"""Lists hook."""
from typing import Union

from tackle import BaseHook, Field


class ListAppendHook(BaseHook):
    """Hook for updating dict objects with items."""

    hook_type: str = 'append'
    input: list = Field(..., description="A list append to.")
    item: Union[list, str] = Field(
        ..., description="A list or string to append to `input` list."
    )

    _args: list = ['input', 'item']

    def execute(self):
        if isinstance(self.item, list):
            for i in self.item:
                self.input.append(i)
        else:
            self.input.append(self.item)

        return self.input


class ListRemoveHook(BaseHook):
    """
    Hook  for updating dict objects with items.

    :param input: A list append to
    :param item: A list or string to remove to `input` list
    :param filter: A regex to remove items from list with
    :return: A list without removed objects objects.
    """

    hook_type: str = 'list_remove'
    input: list = Field(description="A list to append to.")
    item: Union[list, str] = Field(
        ..., description="A list or string to append to `input` list."
    )

    _args: list = ['input', 'item']

    def execute(self):
        if self.item:
            self.input = [i for i in self.input if i != self.item]
        return self.input


class ListFromDictHook(BaseHook):
    """
    Hook generating a list from a dict based on the keys.

    :param input: A list append to
    :param item: A list or string to append to `input` list
    :return: An appended list object.
    """

    hook_type: str = 'list_from_dict'
    keys: list
    input: dict

    def execute(self):
        return [k for k, _ in self.input.items() if k in self.keys]


class ConcatListsHook(BaseHook):
    """
    Hook to concatenate a list of items.

    :param inputs: A list append to
    :param item: A list or string to append to `input` list
    :return: An appended list object.
    """

    hook_type: str = 'concat'
    input: list

    def execute(self):
        output = None
        for i in self.input:
            output = output + i
        return output



"""
Utils for handling command arguments along with unpacking hook arguments. Used in cli
to unpack input args/kwargs and in parser to unpack hook calls -
i.e. key->: hook arg --kwarg thing
"""
import shlex


def strip_dashes(raw_arg: str) -> str:
    """Remove dashes prepended to string for arg splitting."""
    while raw_arg.startswith('-'):
        raw_arg = raw_arg[1:]
    return raw_arg


def unpack_input_string(input_string: str) -> (list, dict, list):
    """Unpack args and"""
    args, kwargs, flags = unpack_args_kwargs_string(input_string)
    if '{{' in args[0]:
        # We split up the string before based on whitespace so eval individually
        if '}}' in args[0]:
            # This is single templatable string -> key->: "{{this}}" => args: ['this']
            args.insert(0, 'var')
        else:
            # Situation where we have key->: "{{ this }}" => args: ['{{', 'this' '}}']
            for i in range(1, len(args)):
                if '}}' in args[i]:
                    joined_template = ' '.join(args[: (i + 1)])
                    other_args = args[(i + 1) :]
                    args = ['var'] + [joined_template] + other_args
                    break

    return args, kwargs, flags


def unpack_args_kwargs_string(input_string: str) -> (list, dict, list):
    """Split up based on whitespace input args and pass to unpack_args_kwargs_list."""
    input_list = shlex.split(input_string)
    return unpack_args_kwargs_list(input_list)


def unpack_args_kwargs_list(input_list: list) -> (list, dict, list):
    """
    Take the input template and unpack the args and kwargs if they exist.
    Updates the command_args and command_kwargs with a list of strings and
    list of dicts respectively.
    """
    input_list_length = len(input_list)
    args = []
    kwargs = {}
    flags = []

    i = 0
    while i < input_list_length:
        raw_arg = input_list[i]
        if i + 1 < input_list_length:
            next_raw_arg = input_list[i + 1]
        else:
            # Allows logic for if last item has `--` in it then it is a flag
            next_raw_arg = "-"

        if (
            raw_arg.startswith('--') or raw_arg.startswith('-')
        ) and not next_raw_arg.startswith('-'):
            # Field is a kwarg
            kwargs.update({strip_dashes(raw_arg): input_list[i + 1]})
            i += 1
        elif (
            raw_arg.startswith('--') or raw_arg.startswith('-')
        ) and next_raw_arg.startswith('-'):
            # Field is a flag
            flags.append(strip_dashes(raw_arg))
        else:
            # Field is an argument
            args.append(strip_dashes(raw_arg))
        i += 1
    return args, kwargs, flags



"""Main parsing module for walking down arbitrary data structures and executing hooks."""
from __future__ import print_function
import logging
from pathlib import Path
import os
import inspect
import warnings
from typing import Type, Any
from pydantic.main import ModelMetaclass, ValidationError

from tackle.providers import import_with_fallback_install
from tackle.render import render_variable, wrap_jinja_braces
from tackle.utils.dicts import (
    nested_get,
    nested_delete,
    nested_set,
    encode_list_index,
    set_key,
    get_readable_key_path,
)
from tackle.utils.command import unpack_args_kwargs_string, unpack_input_string
from tackle.utils.vcs import get_repo_source
from tackle.utils.files import read_config_file
from tackle.utils.paths import (
    work_in,
    is_repo_url,
    is_directory_with_tackle,
    is_file,
    find_tackle_file,
    find_nearest_tackle_file,
)
from tackle.utils.zipfile import unzip
from tackle.models import Context, BaseHook
from tackle.exceptions import (
    HookCallException,
    UnknownHookTypeException,
    UnknownArgumentException,
    UnknownSourceException,
    EmptyTackleFileException,
    EmptyBlockException,
)
from tackle.settings import settings

logger = logging.getLogger(__name__)


def get_hook(hook_type, context: 'Context', suppress_error: bool = False):
    """
    Get the hook from available providers.

    This function does the following to return the hook:
    1. Check if hook hasn't been imported already
    2. Check if the hook has been declared in a provider's __init__.py's
    `hook_types` field.
    3. Try to import it then fall back on installing the requirements.txt file
    """
    for h in BaseHook.__subclasses__():
        if hook_type == inspect.signature(h).parameters['hook_type'].default:
            return h

    for p in context.providers:
        if hook_type in p.hook_types:
            import_with_fallback_install(p.name, p.path)

    for h in BaseHook.__subclasses__():
        if hook_type == inspect.signature(h).parameters['hook_type'].default:
            return h

    avail_hook_types = [
        inspect.signature(i).parameters['hook_type'].default
        for i in BaseHook.__subclasses__()
    ]
    logger.debug(f"Available hook types = {avail_hook_types}")
    if not suppress_error:
        raise UnknownHookTypeException(
            f"The hook type=\"{hook_type}\" is not available in the providers. "
            f"Run the application with `--verbose` to see available hook types."
        )


def evaluate_for(hook_dict: dict, Hook: ModelMetaclass, context: 'Context'):
    """Run the parse_hook function in a loop and return a list of outputs."""
    loop_targets = render_variable(context, wrap_jinja_braces(hook_dict['for']))
    hook_dict.pop('for')

    # Need add an empty list in the value so we have something to append to
    set_key(
        element=context.output_dict,
        keys=context.key_path,
        value=[],
    )

    if len(loop_targets) == 0:
        return

    for i, l in (
        enumerate(loop_targets)
        if not render_variable(context, hook_dict.get('reverse', None))
        else reversed(list(enumerate(loop_targets)))
    ):
        # Create temporary variables in the context to be used in the loop.
        context.existing_context.update({'index': i, 'item': l})
        # Append the index to the keypath
        context.key_path.append(encode_list_index(i))

        # TODO: Do we need to parse a copy of the hook?
        parse_hook(hook_dict.copy(), Hook, context, append_hook_value=True)
        context.key_path.pop()

    # Remove temp variables
    context.existing_context.pop('item')
    context.existing_context.pop('index')


def evaluate_if(hook_dict: dict, context: 'Context', append_hook_value: bool) -> bool:
    """Evaluate the when condition and return bool."""
    if hook_dict.get('for', None) is not None and not append_hook_value:
        # We qualify `if` conditions within for loop logic
        return True
    if hook_dict.get('if', None) is None:
        return True
    return render_variable(context, wrap_jinja_braces(hook_dict['if']))


def evaluate_merge(
    hook_output_value, context: 'Context', append_hook_value: bool = False
):
    """Merge the contents into it's top level set of keys."""
    if append_hook_value:
        raise HookCallException("Can't merge from for loop.")

    if context.key_path[-1] in ('->', '_>'):
        # Expanded key - Remove parent key from key path
        key_path = context.key_path[:-2] + [context.key_path[-1]]
    else:
        # Compact key
        key_path = context.key_path[:-1] + [context.key_path[-1][:-2]]

    # Can't merge into top level keys without merging k/v individually
    if len(key_path) == 1:
        # This is only valid for dict output
        if isinstance(hook_output_value, dict):
            for k, v in hook_output_value.items():
                set_key(
                    element=context.output_dict,
                    keys=[k] + key_path,
                    value=v,
                )
            return
        else:
            raise HookCallException("Can't merge non maps into top level keys.")
    else:
        set_key(
            element=context.output_dict,
            keys=key_path,
            value=hook_output_value,
        )


def render_hook_vars(hook_dict: dict, Hook: ModelMetaclass, context: 'Context'):
    """Render the hook variables."""
    for key, value in list(hook_dict.items()):
        if key in Hook.Config.alias_to_fields:
            # Skip any keys used in logic as these are evaluated / rendered separately
            continue

        if key in Hook._render_exclude or key in Hook._render_exclude_default:
            # Skip anything that has been marked excluded. Needed for things like block
            # hooks which will have templating within the inputs
            continue

        if isinstance(value, str):
            # Check Hook private vars for rendering by default which wraps bare strings
            if key in Hook._render_by_default:
                hook_dict[key] = render_variable(context, wrap_jinja_braces(value))

            # TODO: When we build our own custom Field function then this will change
            elif 'render_by_default' in Hook.__fields__[key].field_info.extra:
                hook_dict[key] = render_variable(context, wrap_jinja_braces(value))

            elif '{{' in value and '}}' in value:
                hook_dict[key] = render_variable(context, value)

        elif isinstance(value, (list, dict)):
            hook_dict[key] = render_variable(context, value)


def parse_hook(
    hook_dict, Hook: ModelMetaclass, context: 'Context', append_hook_value: bool = None
):
    """Parse input dict for loop and when logic and calls hooks."""
    if evaluate_if(hook_dict, context, append_hook_value):

        if 'for' in hook_dict:
            # This runs the current function in a loop with `append_hook_value` set so
            # that keys are appended in the loop.
            evaluate_for(hook_dict, Hook, context)
            return

        else:
            # Render the remaining hook variables
            render_hook_vars(hook_dict, Hook, context)

            try:
                hook = Hook(
                    **hook_dict,
                    input_dict=context.input_dict,
                    output_dict=context.output_dict,
                    existing_context=context.existing_context,
                    no_input=context.no_input,
                    calling_directory=context.calling_directory,
                    providers=context.providers,
                    key_path=context.key_path,
                )
            except ValidationError as e:
                raise e

            # Normal hook run
            hook_output_value = hook.call()

            if hook.merge:
                evaluate_merge(hook_output_value, context, append_hook_value)
            else:
                set_key(
                    element=context.output_dict,
                    keys=context.key_path,
                    value=hook_output_value,
                    append_hook_value=append_hook_value,
                )

    elif 'else' in hook_dict:
        set_key(
            element=context.output_dict,
            keys=context.key_path,
            value=render_variable(context, wrap_jinja_braces(hook_dict['else'])),
            append_hook_value=append_hook_value,
        )


def evaluate_args(args: list, hook_dict: dict, Hook: Type[BaseHook]):
    """
    Associate hook arguments provided in the call with hook attributes. Parses the
    hook's `_args` attribute to know how to map arguments are mapped to where and
    deal with rendering by default.
    """
    for i, v in enumerate(args):
        # Iterate over the input args
        if i + 1 == len(Hook._args):
            # We are at the last argument mapping so we need to join the remaining
            # arguments as a single string if it is not a list of another map.
            # TODO: Validate that we can put Any here as it is too open
            if Hook.__fields__[Hook._args[i]].type_ in (str, float, int, bool, Any):
                # Was parsed on spaces so reconstructed.
                value = ' '.join(args[i:])
            # fmt: skip
            elif isinstance(Hook.__fields__[Hook._args[i]], list):
                # If list then all the remaining items
                value = args[i:]
            else:
                # Only thing left is a dict
                if len(args[i:]) > 1:
                    raise ValueError(
                        f"Can't specify multiple arguments for map argument "
                        f"{Hook.__fields__[Hook._args[i]]}."
                    )
                # Join everything up as a list as it doesn't make sense to do anything
                # else at this point.
                value = args[i]

            hook_dict[Hook._args[i]] = value
            return
        else:
            # The hooks arguments are indexed
            try:
                hook_dict[Hook._args[i]] = v
            except IndexError:
                raise UnknownArgumentException(f"Unknown argument {Hook._args[i]}.")


def run_hook(context: 'Context'):
    """
    Run the hook by qualifying the input argument and matching the input params with the
    with the hook's `_args` which are then overlayed into a hook kwargs. Also interprets
    special cases where you have a string or list input of renderable variables.
    """
    if isinstance(context.input_string, str):
        args, kwargs, flags = unpack_input_string(context.input_string)
        first_arg = args[0]
        # Remove first args it will be consumed and no longer relevant
        args.pop(0)

    else:
        # Rare case when an arrow is used to indicate rendering of a list.
        # Only qualified when input is of form `key->: [{{var}},{{var}},...]
        # In this case we need to set the key as an empty list
        nested_set(
            element=context.output_dict,
            keys=context.key_path[:-1] + [context.key_path[-1][:-2]],
            value=[],
        )
        # Iterate over values appending rendered values. Rendered values can be any type
        for i, v in enumerate(context.input_string):
            nested_set(
                element=context.output_dict,
                keys=context.key_path[:-1]
                + [context.key_path[-1][:-2]]
                + [encode_list_index(i)],
                value=render_variable(context, v),
            )
        return

    # Look up the hook from the imported providers
    Hook = get_hook(first_arg, context, suppress_error=True)

    if Hook is None:
        raise UnknownHookTypeException

    if context.key_path[-1] in ('->', '_>'):
        # We have a expanded or mixed (with args) hook expression and so there will be
        # additional properties in adjacent keys
        hook_dict = nested_get(context.input_dict, context.key_path[:-1]).copy()

        # Need to replace arrow keys as for the time being (pydantic 1.8.2) - multiple
        # aliases for the same field (type) can't be specified so doing this hack
        if '->' in hook_dict:
            hook_dict['hook_type'] = first_arg
            hook_dict.pop('->')
        else:
            hook_dict['hook_type'] = first_arg
            hook_dict.pop('_>')

    else:
        # Hook is a compact expression - Can only be a string
        hook_dict = {}
        # hook_dict['hook_type'] = nested_get(context.input_dict, context.key_path)
        hook_dict['hook_type'] = first_arg

    # Associate hook arguments provided in the call with hook attributes
    evaluate_args(args, hook_dict, Hook)
    # Add any kwargs
    for k, v in kwargs.items():
        hook_dict[k] = v

    # Main parser
    parse_hook(hook_dict, Hook, context)


def handle_empty_blocks(context: 'Context', block_value):
    """
    Handle keys appended with arrows and interpret them as `block` hooks. Value is
    re-written over with a `block` hook to support the following syntax.

    a-key->:
      if: stuff == 'things'
      for: a_list
      foo->: print ...
      bar->: print ...

    to

    a-key:
      ->: block
      if: stuff == 'things'
      for: a_list
      items:
        foo->: print ...
        bar->: print ...

    :param context:
    :param block_key:
    :param block_value:
    :return:
    """
    # Break up key paths
    base_key_path = context.key_path[:-1]
    new_key = [context.key_path[-1][:-2]]

    # Over-write the input with an expanded path (ie no arrow in key)
    nested_set(
        element=context.input_dict,
        keys=base_key_path + new_key,
        value=block_value,
    )
    # Add back the arrow with the value set to `block` for the block hook
    arrow = [context.key_path[-1][-2:]]
    nested_set(
        element=context.input_dict,
        keys=base_key_path + new_key + arrow,
        value='block',
    )
    # Remove the old key
    nested_delete(context.input_dict, context.key_path)

    # Iterate through the block keys except for the reserved keys like `for` or `if`
    aliases = [v.alias for _, v in BaseHook.__fields__.items()] + ['->', '_>']
    for k, v in block_value.copy().items():
        if k not in aliases:
            # Set the keys under the `items` key per the block hook's input
            nested_set(
                element=context.input_dict,
                keys=base_key_path + new_key + ['items', k],
                value=v,
            )
            # Remove the old key
            nested_delete(context.input_dict, base_key_path + new_key + [k])
        elif context.verbose:
            warnings.warn(f"Warning - skipping over {k} in block hook.")

    # Finally check if the `items` key exists in the input_dict.  If not then we have
    # an empty hook which will cause an ambiguous ValidationError for missing field
    if 'items' not in nested_get(context.input_dict, base_key_path + new_key):
        key = get_readable_key_path(base_key_path + new_key)
        raise EmptyBlockException(f"Empty hook in key path = {key}")


def walk_sync(context: 'Context', element):
    """
    Traverse an object looking for hook calls and running those hooks. Here we are
    keeping track of which keys are traversed in a list called `key_path` with strings
    as dict keys and byte encoded integers for list indexes.
    """
    if len(context.key_path) != 0:
        # Handle compact expressions - ie key->: hook_type args
        if context.key_path[-1][-2:] in ('->', '_>'):
            context.input_string = element
            run_hook(context)
            if context.key_path[-1][-2:] == '_>':
                # Private hook calls
                context.keys_to_remove.append(
                    context.key_path[:-1] + [context.key_path[-1][:-2]]
                )
            return

    if isinstance(element, dict):
        # Handle expanded expressions - ie key:\n  ->: hook_type args
        if '->' in element.keys():
            # Public hook calls
            context.key_path.append('->')
            context.input_string = element['->']
            run_hook(context)
            context.key_path.pop()
            return
        elif '_>' in element.keys():
            # Private hook calls
            context.key_path.append('_>')
            context.input_string = element['_>']
            run_hook(context)
            context.key_path.pop()
            context.keys_to_remove.append(context.key_path.copy())
            return

        for k, v in element.copy().items():
            context.key_path.append(k)
            # Special case where we have an empty hook, expanded or compact
            if k[-2:] in ('->', '_>') and (v is None or isinstance(v, dict)):
                # Here we re-write the input to turn empty hooks into block hooks
                handle_empty_blocks(context, v)
                context.key_path[-1] = k[:-2]
                walk_sync(context, v)
                context.key_path.pop()
            else:
                # Recurse
                walk_sync(context, v)
                context.key_path.pop()

    # Non-hook calls recurse through inputs
    elif isinstance(element, list):
        for i, v in enumerate(element.copy()):
            context.key_path.append(encode_list_index(i))
            walk_sync(context, v)
            context.key_path.pop()
    else:
        nested_set(element=context.output_dict, keys=context.key_path, value=element)


def run_handler(context, handler_key, handler_value):
    """
    Run a pre/post execution handlers which are either hooks or functions.

    NOTE: This is an experimental feature and may change.
    """
    if handler_key in context.functions:
        """Run functions"""
        function = context.functions[handler_key]
        context.input_dict = function.exec
        walk_sync(context, function.exec.copy())

    elif get_hook(handler_key, context, suppress_error=True):
        Hook = get_hook(handler_key, context)
        # context.input_string('import ' + handler_value)
        # run_hook_function(context)
        # if isinstance()
        hook = Hook(
            **handler_value,
            input_dict=context.input_dict,
            output_dict=context.output_dict,
            no_input=context.no_input,
            providers=context.providers,
        )
        hook.call()
    else:
        raise UnknownHookTypeException()


def update_input_dict_with_kwargs(context: 'Context', kwargs: dict):
    """
    Update the input dict with kwargs which in this context are treated as overriding
    the keys. Takes into account if the key is a hook and replaces that.
    """
    for k, v in kwargs.items():
        if k in context.input_dict:
            context.input_dict.update({k: v})
        elif f"{k}->" in context.input_dict:
            context.input_dict.pop(f"{k}->")
            context.input_dict.update({k: v})


def run_source(context: 'Context', args: list, kwargs: dict, flags: list):
    """
    Take the input dict and impose global args/kwargs/flags with the following logic:
    - Use kwargs/flags as overriding keys in the input_dict
    - Check the input dict if there is a key matching the arg and run that key
      - Additional arguments are assessed as
        - If the call is to a hook directly, then inject that as an argument
        - If the call is to a block of hooks then call the next hook key
    - Otherwise run normally (ie full parsing).

    An exception exists for if the last arg is `help` in which case that level's help
    is called and exited 0.
    """
    # Tackle is called both through the CLI and as a package and so to preserve args /
    # kwargs we
    if context.global_args is not None:
        args = args + context.global_args
        context.global_args = None

    # Global kwargs/flags are immediately consumed and injected into the kwargs/flags
    if context.global_kwargs is not None:
        kwargs.update(context.global_kwargs)
        context.global_kwargs = None

    if context.global_flags is not None:
        kwargs.update({i: True for i in context.global_flags})
        context.global_kwargs = None

    update_input_dict_with_kwargs(context=context, kwargs=kwargs)

    for i in flags:
        # Process flags by setting key to true
        context.input_dict.update({i: True})

    if len(args) >= 1:
        # TODO: Implement help
        # `help` which will always be the last arg
        # if args[-1] == 'help':
        #     # Calling help will exit 0. End of the line.
        #     run_help(context, context.input_dict, args[:-1])

        # Loop through all args
        for i in args:
            # Remove any arrows on the first level keys
            first_level_compact_keys = [
                k[:-2] for k, _ in context.input_dict.items() if k.endswith('->')
            ]
            if i in first_level_compact_keys:
                arg_key_value = context.input_dict[i + '->']
                if isinstance(arg_key_value, str):
                    # We have a compact hook so nothing else to traverse
                    break

            elif i in context.input_dict:
                context.key_path.append(i)
                walk_sync(context, context.input_dict[i].copy())
                context.key_path.pop()
            else:
                raise ValueError(f"Argument {i} not found as key in input.")
        return
    walk_sync(context, context.input_dict.copy())


def extract_base_file(context: 'Context'):
    """Read the tackle file and initialize input_dict."""
    path = os.path.join(context.input_dir, context.input_file)
    input_dict = read_config_file(path)
    context.input_dict = input_dict

    if input_dict is None:
        raise EmptyTackleFileException(f"No tackle file found at {path}.")

    # TODO: Experimental feature that could be integrated later
    # # Extract handlers
    # for k, v in list(input_dict.items()):
    #     if k.startswith('__'):
    #         # Run pre-execution handlers and remove from input
    #         run_handler(context, k[2:], v)
    #         input_dict.pop(k)
    #     if k.endswith('__'):
    #         # Store post-execution handlers and remove from input
    #         # TODO: Execute post exec handlers
    #         context.post_exec_handlers.append({k[:-2]: v})
    #         input_dict.pop(k)

    context.input_dict = input_dict


def import_local_provider_source(context: 'Context', provider_dir: str):
    """
    Import a provider from a path by checking if the provider has a tackle file and
    returning a path.
    """
    context.input_dir = provider_dir
    if context.input_file is None:
        context.input_file = find_tackle_file(provider_dir)

    if context.directory:
        context.input_file = os.path.join(context.input_file, context.directory)

    extract_base_file(context)


def update_source(context: 'Context'):
    """
    Locate the repository directory from a template reference. This is the main parser
    for determining the source of the context and calls the succeeding parsing
    functions. The parsing order has the following order of precedence.

    If the template wasn't given then use the file in that parent directory.
    If the template refers to a zip file or zip url, download / unzip as the context.
    If the template refers to a repository URL, clone it.
    If the template refers to a file, use that as the context.
    """
    args, kwargs, flags = unpack_args_kwargs_string(context.input_string)

    first_arg = args[0]
    # Remove first args it will be consumed and no longer relevant
    args.pop(0)

    # Zipfile
    if first_arg.lower().endswith('.zip'):
        unzipped_dir = unzip(
            zip_uri=first_arg,
            clone_to_dir=settings.tackle_dir,
            no_input=context.no_input,
            password=context.password,  # TODO: RM - Should prompt?
        )
        import_local_provider_source(context, unzipped_dir)
    # Repo
    elif is_repo_url(first_arg):
        provider_dir = get_repo_source(first_arg, context.version)
        import_local_provider_source(context, provider_dir)
    # Directory
    elif is_directory_with_tackle(first_arg):
        # Special case where the input is a path to a directory. Need to override some
        # settings that would normally get populated by zip / repo refs
        context.input_file = os.path.basename(find_tackle_file(first_arg))
        context.input_dir = Path(first_arg).absolute()

        # Load the base file into input_dict
        extract_base_file(context)
    # File
    elif is_file(first_arg):
        context.input_file = os.path.basename(first_arg)
        context.input_dir = Path(first_arg).parent.absolute()
        extract_base_file(context)
    # Search in parent
    else:
        # Lastly we check if there is a key in the parent that matches the arg. This is
        # basically the fallback logic but need to raise error right away if the key
        # does not exist so we don't have to catch it with context later.
        tackle_file = find_nearest_tackle_file()
        if tackle_file is None:
            raise UnknownSourceException(f"Could not find source = {first_arg}")

        context.input_file = os.path.basename(tackle_file)
        context.input_dir = Path(tackle_file).parent.absolute()
        extract_base_file(context)

        if first_arg not in context.input_dict:
            raise UnknownSourceException(
                f"Could not find source = {first_arg} or as "
                f"key in parent tackle file."
            )
        args.insert(0, first_arg)

    # We always change directory into the source that is being called. Needs to be this
    # or would be very confusing if writing a provider to always refer to it's own path.
    with work_in(context.input_dir):
        # Main parsing logic
        run_source(context, args, kwargs, flags)



"""Random number / string hooks."""
import random
import string
import logging

from tackle.models import BaseHook, Field
from pydantic import validator
from tackle.exceptions import HookCallException

logger = logging.getLogger(__name__)


class RandomHexHook(BaseHook):
    """Hook  for `random_hex`. Lists the contents of a directory."""

    hook_type: str = 'random_hex'
    length: int = Field(8, description="Number for number of digits - default 8")

    _args: list = ['length']

    def execute(self) -> str:
        return ''.join(['%0', str(self.length), 'x']) % random.randrange(
            16 ** self.length
        )


class RandomStringHook(BaseHook):
    """Hook  for `random_string`. Lists the contents of a directory."""

    hook_type: str = 'random_string'
    string_case: str = Field(
        'lower', description="Case of output, one of `upper` or `lower`"
    )
    length: int = Field(8, description="Length of the random string.")

    _args: list = ['length', 'string_case']

    @validator('string_case', allow_reuse=True)
    def validate_string_case2(cls, v):
        if v not in ['upper', 'lower']:
            raise HookCallException(
                "Field `string_case` must be one of 'upper' or 'lower'."
            )
        return v

    def execute(self):
        if self.string_case == 'upper':
            return ''.join(
                random.choices(string.ascii_uppercase + string.digits, k=self.length)
            )
        elif self.string_case == 'lower':
            return ''.join(
                random.choices(string.ascii_lowercase + string.digits, k=self.length)
            )



"""Checkbox hook."""
import logging
from PyInquirer import prompt
from typing import Union, List, Any, Dict
from pydantic import Field

from tackle.models import BaseHook
from tackle.utils.dicts import get_readable_key_path

logger = logging.getLogger(__name__)


class InquirerCheckboxHook(BaseHook):
    """
    Hook for PyInquirer type prompts.

    Takes in three forms of `choices` inputs.
    1. list of string
    2. list of maps with all keys = `name`
    3. list of maps with the key as the output, the value as displayed question

    :param message: String message to show when prompting.
    :param choices: A list of strings or list of k/v pairs per above description
    :param name: A key to insert the output value to. If not provided defaults to
        inserting into parent key
    :param qmark: A marker to select with like 😃
    :return: List of answers
    """

    hook_type: str = 'checkbox'

    message: str = Field(None, description="String message to show when prompting.")
    index: bool = Field(
        False, description="Boolean to return the output in a list."
    )  # TODO: Fix this?
    default: Any = Field([], description="Default for the return value")
    choices: Union[List[str], List[Dict]] = Field(
        ..., description="Either a list of strings or dictionary ."
    )
    checked: bool = Field(
        False, description="Boolean if the default choices should all be checked."
    )
    name: str = 'tmp'

    _args: list = ['message']

    def __init__(self, **data: Any):
        super().__init__(**data)
        if self.message is None:
            self.message = get_readable_key_path(self.key_path) + ' >>>'

    def execute(self) -> list:
        if self.no_input:
            return self.default

        choices_type = None
        for i, v in enumerate(self.choices):
            if i != 0 and type(v) != choices_type:
                raise ValueError("All items need to be of the same type")
            choices_type = type(v)

        if choices_type == str:
            self.choices = [
                {'name': x} if isinstance(x, str) else x for x in self.choices
            ]

            answer = self._run_prompt()
            # if self.index:
            #     return self.choices.index(answer)
            # else:
            return answer

        elif choices_type == dict:
            # This is the normal input to the Hook ie
            # choices = ['name': 'stuff', 'name': 'things']
            keys = list(list(i.keys())[0] for i in self.choices)
            num_keys = list(dict.fromkeys(keys))
            if len(num_keys) == 1 and keys[0] == 'name':
                return self._run_prompt()

            # Otherwise we expect to reindex the key as the output per this:
            # choices = ['stuff': 'How much stuff?', 'things': 'How many things?']
            choices = []
            for i in self.choices:
                choices.append(i[list(i.keys())[0]])

            # Fixing to the expected input choices {'name': 'stuff', 'name': ...}
            choices_dict = self.choices
            self.choices = [{'name': x} if isinstance(x, str) else x for x in choices]

            answer = self._run_prompt()

            # This is for reindexing the output
            output = []
            for i, v in enumerate(choices_dict):
                val = list(v.keys())[0]
                if v[val] in answer:
                    if self.index:
                        output.append(i)
                    else:
                        output.append(val)
            return output

    def _run_prompt(self):
        question = {
            'type': 'checkbox',
            'name': self.name,
            'message': self.message,
            'choices': self.choices,
            'checked': self.checked,
        }
        if self.default:
            question.update({'default': self.default})
        response = prompt([question])

        if self.name != 'tmp':
            return response
        else:
            return response['tmp']



"""
Utils for modifying complex dictionaries generally based on an encoded key_path which is
a list of strings for key value lookups and byte encoded integers for items in a list.
"""


def encode_list_index(list_index: int) -> bytes:
    """Encode an index for lists in a key path to bytes."""
    return list_index.to_bytes(2, byteorder='big')


def decode_list_index(list_index: bytes) -> int:
    """Decode an index for lists in a key path from bytes."""
    return int.from_bytes(list_index, byteorder='big')


def remove_dead_leaves():
    """Traverse up a generic element and removing empty elements."""
    # TODO: Implement?
    pass


def get_key_from_key_path(key_path: list) -> str:
    """Take key_path and return the nearest key from end removing arrows."""
    if key_path[-1] in ('->', '_>'):
        # Expanded key
        return key_path[-2]
    elif key_path[-1].endswith('->', '_>'):
        # Compact key
        return key_path[-1][:-2]


def get_readable_key_path(key_path: list) -> str:
    """Take key_path and return the nearest key from end removing arrows."""
    readable_key_path = []
    for i in key_path:
        if i in ('->', '_>'):
            return '.'.join(readable_key_path)
        elif i[-2:] in ('->', '_>'):
            readable_key_path.append(i[:-2])
            return '.'.join(readable_key_path)
        if isinstance(i, bytes):
            readable_key_path.append(decode_list_index(i))
        else:
            readable_key_path.append(i)


def nested_delete(element, keys):
    """
    Delete items in a generic element (list / dict) based on a key path in the form of
    a list with strings for keys and byte encoded integers for indexes in a list.
    """
    num_elements = len(keys)

    if num_elements == 1:
        if isinstance(keys[0], bytes):
            element.pop(decode_list_index(keys[0]))
            return element

        elif isinstance(element, dict):
            element.pop(keys[0])
            return

    elif num_elements == 2:
        if isinstance(keys[0], bytes) and isinstance(keys[1], str):
            # Case for when we have an embedded item in a list but don't know if it is
            # a map with multiple keys that needs to have a single key removed or the
            # whole item itself
            # TODO
            if isinstance(element[decode_list_index(keys[0])], dict):
                # Further inspection to see if there are any other keys in the dict.
                # If not, then remove the whole item otherwise recurse
                if len(element[decode_list_index(keys[0])].keys()) == 1:
                    element.pop(decode_list_index(keys[0]))
                    return

                return nested_delete(element[decode_list_index(keys[0])], keys[1:])
        elif isinstance(keys[0], str) and isinstance(keys[1], bytes):
            return nested_delete(element[keys[0]], keys[1:])

        elif isinstance(keys[0], str) and isinstance(keys[1], str):
            #
            if isinstance(element[keys[0]], dict):
                # Case where we have an embedded
                return nested_delete(element[keys[0]], keys[1:])
            else:
                # element.pop(keys[0])
                element[keys[0]] = None
                return

    if isinstance(keys[0], bytes):
        return nested_delete(element[decode_list_index(keys[0])], keys[1:])

    return nested_delete(element[keys[0]], keys[1:])


def nested_get(element, keys):
    """
    Getter for dictionary / list elements based on a key_path.

    :param element: A generic dictionary or list
    :param keys: List of string and byte encoded integers.
    :return: The value from the key_path
    """
    num_elements = len(keys)

    if num_elements == 1:
        if isinstance(keys[0], bytes):
            return element[decode_list_index(keys[0])]
        return element[keys[0]]

    if isinstance(keys[0], bytes):
        return nested_get(element[decode_list_index(keys[0])], keys[1:])
    return nested_get(element[keys[0]], keys[1:])


def nested_set(element, keys, value, index: int = 0):
    """
    Set the value of an arbitrary object based on a key_path in the form of a list
    with strings for keys and byte encoded integers for indexes in a list. This function
    recurses through the element until it is at the end of the keys where it sets it.

    :param element: A generic dictionary or list
    :param keys: List of string and byte encoded integers.
    :param value: The value to set
    :param index: Index is only used when called recursively, not initially
    """
    num_elements = len(keys)
    # Check if we are at the last element of the list to insert the value
    if index == num_elements - 1:
        if isinstance(keys[-1], bytes):
            element.insert(decode_list_index(keys[-1]), value)
        else:
            element[keys[-1]] = value
        return

    # Check if the element is a dict and
    if isinstance(element, dict):
        # Look ahead if the next item is a list
        if isinstance(keys[index + 1], bytes):
            element = element.setdefault(keys[index], [])
        else:
            element = element.setdefault(keys[index], {})

    # element must be list
    # Check if we are updating an existing element
    elif len(element) > decode_list_index(keys[index]):
        element = element[decode_list_index(keys[index])]
    else:
        if isinstance(keys[index + 1], bytes):
            element.insert(decode_list_index(keys[index]), [])
        else:
            element.insert(decode_list_index(keys[index]), {})
        element = element[decode_list_index(keys[index])]
    nested_set(element, keys, value, index + 1)


def set_key(element, keys: list, value, append_hook_value: bool = False):
    """
    Wrap nested_set to set keys for both public and private hook calls.

    For public hook calls, qualifies if the hook is compact form (ie key->) or expanded
    (ie key: {->:..}) before setting the output. For private hook calls, the key and
    all parent keys without additional objects are deleted later as they might be
    used in rendering so they are added as well but their key paths are tracked for
    later deletion.
    """
    # TODO: Implement removal of parent keys without values
    if isinstance(keys[-1], bytes):
        if append_hook_value:
            # Condition when we are appending values from a hook for loop where we need
            # take off the prior keys hook arrows
            parent_key_value = keys[-2]
            if parent_key_value in ('->', '_>'):
                # We never append with the hook arrow so temporarily remove it so we can
                # still keep the path to the key
                keys.pop(-2)
                nested_set(element, keys, value)
                keys.insert(-1, parent_key_value)
            else:
                # Same as above but here we are in a compact expression and want to
                # temporarily replace the trailing hook arrow
                keys[-2] = parent_key_value.replace('->', '').replace('_>', '')
                nested_set(element, keys, value)
                keys[-2] = parent_key_value
        else:
            # Condition when we are appending values from a list
            nested_set(element, keys, value)

    elif keys[-1] == '->':  # Expanded public hook call
        nested_set(element, keys[:-1], value)
    elif keys[-1].endswith('->'):  # Compact public hook call
        nested_set(element, keys[:-1] + [keys[-1][:-2]], value)
    elif keys[-1] == '_>':  # Expanded private hook call
        nested_set(element, keys[:-1], value)
    elif keys[-1].endswith('_>'):  # Compact private hook call
        key_path = keys[:-1] + [keys[-1][:-2]]
        nested_set(element, key_path, value)



"""JSON hooks."""
import logging
import json
from typing import Union

from tackle.models import BaseHook, Field

logger = logging.getLogger(__name__)


class JsonHook(BaseHook):
    """
    Hook for json. If no `contents` is provided, the hook reads from path. Otherwise
     it writes the `contents`. When writing, returns path. When reading, returns dict.
    """

    hook_type: str = 'json'
    path: str = Field(..., description="The path to write the file.")
    contents: dict = Field(None, description="A dict to write")

    _args: list = ['path', 'contents']

    def execute(self) -> Union[dict, str]:
        if self.contents:
            with open(self.path, 'w') as f:
                json.dump(self.contents, f)
            return self.path

        else:
            with open(self.path, 'r') as f:
                contents = json.load(f)
            return contents



"""Main entrypoint for rendering."""
import ast
import re
from jinja2 import meta
from inspect import signature

from tackle.render.special_vars import special_variables
from tackle.render.environment import StrictEnvironment
from tackle.exceptions import UnknownTemplateVariableException
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from tackle.models import Context


def wrap_braces_if_not_exist(value):
    """Wrap with braces if they don't exist."""
    if '{{' in value and '}}' in value:
        # Already templated
        return value
    return '{{' + value + '}}'


def wrap_jinja_braces(value):
    """Wrap a string with braces so it can be templated."""
    if isinstance(value, str):
        return wrap_braces_if_not_exist(value)
    # Nothing else can be wrapped
    return value


def render_variable(context: 'Context', raw: Any):
    """
    Render the raw input. Does recursion with dict and list inputs, otherwise renders
    string.

    :param raw: The value to be rendered.
    :return: The rendered value as literal type.
    """
    if raw is None:
        return None
    elif isinstance(raw, str):
        render_string(context, raw)
    elif isinstance(raw, dict):
        return {
            render_string(context, k): render_variable(context, v)
            for k, v in raw.items()
        }
    elif isinstance(raw, list):
        return [render_variable(context, v) for v in raw]
    else:
        return raw

    return render_string(context, raw)


def render_string(context: 'Context', raw: str):
    """
    Render strings by first extracting renderable variables, build render context from
    the output_dict, then existing context, and last looks up special variables.

    :param raw: A renderable string
    :return: The literal value if the output is a string / list / dict / float / int
    """
    if '{{' not in raw:
        return raw

    env = StrictEnvironment(context=context.input_dict)
    template = env.from_string(raw)
    # Extract variables
    variables = meta.find_undeclared_variables(env.parse(raw))

    # Build a render context by inspecting the renderable variables
    render_context = {}
    unknown_variable = []
    for v in variables:
        # Variables in the current output_dict take precedence
        if v in context.output_dict:
            render_context.update({v: context.output_dict[v]})
        elif v in context.existing_context:
            render_context.update({v: context.existing_context[v]})
        elif v in special_variables:
            # If it is a special variable we need to check if the call requires
            # arguments, only context supported now.
            argments = list(signature(special_variables[v]).parameters)
            if len(argments) == 0:
                render_context.update({v: special_variables[v]()})
            elif 'context' in argments:
                render_context.update({v: special_variables[v](context)})
            else:
                raise ValueError("This should never happen.")
        else:
            unknown_variable.append(v)

    try:
        rendered_template = template.render(render_context)
    except Exception as e:
        if len(unknown_variable) != 0:
            raise UnknownTemplateVariableException(
                f"Variable {unknown_variable} unknown."
            )
        raise e

    # ast.literal_eval fails on string like objects so qualifying first
    # This might be dumb but works
    REGEX = [
        r'^\[.*\]$',  # List
        r'^\{.*\}$',  # Dict
        r'^True$|^False$',  # Boolean
        r'^\d+$',  # Integer
        r'^[+-]?([0-9]+([.][0-9]*)?|[.][0-9]+)$',  # Float
    ]
    for r in REGEX:
        if bool(re.search(r, rendered_template)):
            return ast.literal_eval(rendered_template)

    return rendered_template



"""Meta hooks."""
from git import Repo
import logging
from tackle.models import BaseHook
from typing import Optional

logger = logging.getLogger(__name__)


class GitCloneHook(BaseHook):
    """
    Hook to create clone a repo.

    Wraps https://gitpython.readthedocs.io/en/stable/reference.html#git.repo.base.Repo.clone

    :param url: – valid git url, see
        http://www.kernel.org/pub/software/scm/git/docs/git-clone.html#URLS
    :param to_path: – Path to which the repository should be cloned to
    :param progress: – See ‘git.remote.Remote.push’.
    :param env: – Optional dictionary containing the desired environment variables.
        Note: Provided variables will be used to update the execution environment for git.
        If some variable is not specified in env and is defined in os.environ, value from
        os.environ will be used. If you want to unset some variable, consider providing empty
        string as its value.
    :param path: – is the full path of the new repo (traditionally ends with ./<name>.git).
    :param progress: – See ‘git.remote.Remote.push’.
    :param multi_options: – A list of Clone options that can be provided multiple times. One
        option per list item which is passed exactly as specified to clone. For example
        [‘–config core.filemode=false’, ‘–config core.ignorecase’, ‘–recurse-submodule=repo1_path’,
        ‘–recurse-submodule=repo2_path’]
    :param kwargs: –
            odbt = ObjectDatabase Type, allowing to determine the object database implementation
                used by the returned Repo instance
            All remaining keyword arguments are given to the git-clone command
    """

    hook_type: str = 'git_clone'
    url: str = None
    to_path: str = None
    progress: Optional[str] = None
    env: Optional[dict] = None
    multi_options: Optional[list] = None
    kwargs: Optional[dict] = {}

    def execute(self):
        if not self.to_path:
            self.to_path = self.url.split('/')[-1].split('.')[0]
        if self.url:
            Repo.clone_from(
                url=self.url,
                to_path=self.to_path,
                progress=self.progress,
                env=self.env,
                multi_options=self.multi_options,
                **self.kwargs,
            )
            return self.to_path



"""Models for the project."""
import os
from pathlib import Path
from pydantic import BaseModel, SecretStr, Field, Extra, validator
from typing import Any, Union, Optional

from tackle.utils.paths import work_in
from tackle.providers import ProviderList
from tackle.render import wrap_jinja_braces


class Context(BaseModel):
    """The main object that is being modified by parsing."""

    # Mode
    no_input: bool = False
    verbose: bool = False

    # Source
    password: SecretStr = None
    directory: str = None

    # Inputs
    input_string: str = None
    input_dir: Path = None
    input_file: str = None
    version: str = Field(
        None,
        description="For inputs referencing remote repos, this refers to "
        "a branch or tag.",
    )
    existing_context: dict = {}
    overwrite_inputs: Union[dict, str] = None

    input_dict: dict = {}
    output_dict: dict = {}
    keys_to_remove: list = []

    # Internal
    key_path: list = []
    providers: ProviderList = None
    calling_directory: Path = None
    env: Any = None

    global_args: list = None
    global_kwargs: dict = None
    global_flags: list = None

    def __init__(self, **data: Any):
        super().__init__(**data)
        # Allows for passing the providers between tackle runtimes
        if self.providers is None:
            # Native and settings.extra_providers initialized
            self.providers = ProviderList()
        if self.calling_directory is None:
            # Can be carried over from another context. Should only be initialized when
            # called from CLI.
            self.calling_directory = os.path.abspath(os.path.curdir)


class BaseHook(BaseModel):
    """
    Base hook class from which all other hooks inherit from to be discovered. There are
    a number of reserved keys that are used for logic such as `if` and `for` that are
    aliased to `if_` and `for_` to not collide with python reserved key words. We also
    append underscores to

    """

    hook_type: str = Field(..., description="Name of the hook.")

    # TODO: Serialize the access modifier earlier in parsing - Could store the arrow in
    #  key_path as boolean. Would have big changes to tackle.utils.dicts
    # access_modifier: Literal['public', 'private'] = None

    if_: Union[str, bool] = Field(None, render_by_default=True)
    else_: Union[str, bool] = Field(None, render_by_default=True)
    for_: Union[str, list] = Field(None, render_by_default=True)
    reverse: Union[str, bool] = Field(None, render_by_default=True)

    callback: str = None
    chdir: Optional[str] = Field(None, description="Name of the hook.")
    merge: Union[bool, str] = None
    confirm: Optional[Any] = None

    # context: Context
    input_dict: Union[dict, list] = None
    output_dict: Union[dict, list] = {}
    existing_context: dict = None
    no_input: bool = None
    calling_directory: Path = None

    providers: ProviderList = None
    key_path: list = None

    # Placeholder until help can be fully worked out
    help: str = None

    _args: list = []
    _kwargs: dict = {}
    _flags: list = []
    # Fields that should not be rendered by default
    _render_exclude_default: set = {'input_dict', 'output_dict', 'hook_type'}
    _render_exclude: set = {}
    _render_by_default: list = []

    @validator('if_', 'else_', 'reverse', 'for_', 'merge')
    def wrap_bool_if_string(cls, v):
        return wrap_jinja_braces(v)

    # Per https://github.com/samuelcolvin/pydantic/issues/1577
    # See below
    def __setattr__(self, key, val):
        """Override method to alias input fields."""
        if key in self.__config__.alias_to_fields:
            key = self.__config__.alias_to_fields[key]
        super().__setattr__(key, val)

    class Config:
        arbitrary_types_allowed = True
        extra = Extra.forbid
        validate_assignment = True
        fields = {
            'if_': 'if',
            'else_': 'else',
            'match_': 'match',
            'case_': 'case',
            'for_': 'for',
            'while_': 'while',
            'enumerate_': 'enumerate',
        }
        # Per https://github.com/samuelcolvin/pydantic/issues/1577
        # This is an issue until pydantic 1.9 is released and items can be set with
        # properties which will override the internal __setattr__ method that
        # disregards aliased fields
        alias_to_fields = {v: k for k, v in fields.items()}

    def __init__(self, **data: Any):
        super().__init__(**data)

    def execute(self) -> Any:
        raise NotImplementedError("Every hook needs an execute method.")

    def call(self) -> Any:
        """
        Call main entrypoint to calling hook.

        Handles `chdir` method.
        """
        if self.chdir:
            path = os.path.abspath(os.path.expanduser(self.chdir))
            if os.path.isdir(path):
                # Use contextlib to switch dirs
                with work_in(os.path.abspath(os.path.expanduser(self.chdir))):
                    return self.execute()
            else:
                raise NotADirectoryError(
                    f"The specified path='{path}' to change to was not found."
                )
        else:
            return self.execute()



# -*- coding: utf-8 -*-

"""Tests dict input objects for `cookiecutter.parser.providers` module."""
import os

import pytest
from tackle.main import tackle
import subprocess
import sys


@pytest.fixture()
def temporary_uninstall(package='art'):
    """Fixture to uninstall a package and install it after the test."""
    subprocess.check_call(
        [
            sys.executable,
            "-m",
            "pip",
            "uninstall",
            "--quiet",
            "--disable-pip-version-check",
            "-y",
            package,
        ]
    )
    yield


def test_parser_provider_import_installs_requirements(
    change_curdir_fixtures, temporary_uninstall
):
    """Validate that if a package is missing, that it will be installed and usable."""
    tackle('text2art.yaml')


def test_parser_hooks_raises_error_on_unknown_hook_type(change_curdir_fixtures):
    """Verify raising error.

    Verify that the hook parser raises the right error when the hook type is
    not in the providers.
    """
    from tackle.exceptions import UnknownHookTypeException

    with pytest.raises(UnknownHookTypeException):
        tackle('unknown-hook-type.yaml')


def test_parser_provider_hook_add(change_curdir_fixtures):
    """Validate adding providers.

    Validate that you can give a `__provider` key to point to
    additional providers and make them available as a type.
    """
    os.chdir('test-provider')
    o = tackle('context_provider.yaml')
    assert o['things'] == 'bar'


def test_parser_provider_hook_add_list(change_curdir_fixtures):
    """Validate adding providers.

    Validate that you can give multiple `__provider` keys to point to
    additional providers and make them available as a types.
    """
    o = tackle('context_provider_list.yaml')
    assert o['things'] == 'bar'
    assert o['stuff'] == 'bar'



"""Test for tackle.utils.command"""
import pytest

from tackle.utils.command import unpack_args_kwargs_string, unpack_input_string

TEMPLATES = [
    ('foo bar baz', 3, 0, 0),
    ('foo --bar baz bing', 2, 1, 0),
    ('foo bar --baz foo', 2, 1, 0),
    ('foo bar --baz foo --bing baz', 2, 2, 0),
    ('foo --bar baz', 1, 1, 0),
    ('foo bar --baz', 2, 0, 1),
    ('foo bar baz --bing', 3, 0, 1),
    ('foo --bar baz --foo', 1, 1, 1),
    ('foo bar --foo bar --bing --baz bling', 2, 2, 1),
    ('foo --bar baz blah --bling', 2, 1, 1),
    ('this --if "expanded == \'that\'"', 1, 1, 0),
    # These should all have `var` prepended to the args as it is indicative of a render
    ('"{{foo}}" bar baz', 4, 0, 0),
    ('{{ foo }} bar baz', 4, 0, 0),
    ('{{ foo }} bar baz bing', 5, 0, 0),
    ('{{ foo}} bar baz', 4, 0, 0),
    ('{{foo }} bar baz', 4, 0, 0),
    ('{{ foo }}-foo bar baz', 4, 0, 0),
    ('bar-{{ foo }}-foo', 2, 0, 0),
    ('bar-{{ foo in var }}-foo', 2, 0, 0),
]


@pytest.mark.parametrize("template,len_args,len_kwargs,len_flags", TEMPLATES)
def test_unpack_args_kwargs(template, len_args, len_kwargs, len_flags):
    """Validate the count of each input arg/kwarg/flag."""
    args, kwargs, flags = unpack_input_string(template)

    assert len_args == len(args)
    assert len_kwargs == len(kwargs)
    assert len_flags == len(flags)


FIXTURES = [
    # input_string, args, kwargs, flags
    ("this --if \"expanded == 'that'\"", ["this"], {"if": "expanded == 'that'"}, []),
    (
        "this that --if \"expanded == 'that'\"",
        ["this", "that"],
        {"if": "expanded == 'that'"},
        [],
    ),
]


@pytest.mark.parametrize("input_string,args,kwargs,flags", FIXTURES)
def test_unpack_input_string(input_string, args, kwargs, flags):
    """Validate expressions for input strings."""
    args_out, kwargs_out, flags_out = unpack_args_kwargs_string(input_string)
    assert args_out == args
    assert kwargs_out == kwargs
    assert flags_out == flags



"""Directory hooks."""
import logging
import pathlib

from tackle.models import BaseHook, Field

logger = logging.getLogger(__name__)


class MakeDirectoryHook(BaseHook):
    """Hook creating a directory."""

    hook_type: str = 'mkdir'
    path: str = Field(..., description="The path to file or directory")

    _args: list = ['path']

    def execute(self) -> str:
        pathlib.Path(self.path).mkdir(parents=True, exist_ok=True)
        return self.path
